from typing import Optional, Dict, List, Any

from phi.k8s.app.base import (
    K8sApp,
    AppVolumeType,  # noqa: F401
    ContainerContext,  # noqa: F401
    ServiceType,
    RestartPolicy,  # noqa: F401
    ImagePullPolicy,  # noqa: F401
    LoadBalancerProvider,  # noqa: F401
)
from phi.k8s.app.traefik.crds import ingressroute_crd, middleware_crd
from phi.utils.log import logger


class TraefikRouter(K8sApp):
    # -*- App Name
    name: str = "traefik"

    # -*- Image Configuration
    image_name: str = "traefik"
    image_tag: str = "v2.10"

    # -*- RBAC Configuration
    # Create a ServiceAccount, ClusterRole, and ClusterRoleBinding
    create_rbac: bool = True

    # -*- Install traefik CRDs
    # See: https://doc.traefik.io/traefik/providers/kubernetes-crd/#configuration-requirements
    install_crds: bool = True

    # -*- Traefik Configuration
    domain_name: Optional[str] = None
    # Enable Access Logs
    access_logs: bool = True
    # Traefik config file on the host
    traefik_config_file: Optional[str] = None
    # Traefik config file on the container
    traefik_config_file_container_path: str = "/etc/traefik/traefik.yaml"

    # -*- HTTP Configuration
    http_enabled: bool = False
    http_routes: Optional[List[dict]] = None
    http_container_port: int = 80
    http_service_port: int = 80
    http_node_port: Optional[int] = None
    http_key: str = "http"
    http_ingress_name: str = "http-ingress"
    forward_http_to_https: bool = False
    enable_http_proxy_protocol: bool = False
    enable_http_forward_headers: bool = False

    # -*- HTTPS Configuration
    https_enabled: bool = False
    https_routes: Optional[List[dict]] = None
    https_container_port: int = 443
    https_service_port: int = 443
    https_node_port: Optional[int] = None
    https_key: str = "https"
    https_ingress_name: str = "https-ingress"
    enable_https_proxy_protocol: bool = False
    enable_https_forward_headers: bool = False
    add_headers: Optional[Dict[str, dict]] = None

    # -*- Dashboard Configuration
    dashboard_enabled: bool = False
    dashboard_routes: Optional[List[dict]] = None
    dashboard_container_port: int = 8080
    dashboard_service_port: int = 8080
    dashboard_node_port: Optional[int] = None
    dashboard_key: str = "dashboard"
    dashboard_ingress_name: str = "dashboard-ingress"
    # The dashboard is gated behind a user:password, which is generated using
    #   htpasswd -nb user password
    # You can provide the "users:password" list as a dashboard_auth_users param
    # or as DASHBOARD_AUTH_USERS in the secrets_file
    # Using the secrets_file is recommended
    dashboard_auth_users: Optional[str] = None
    insecure_api_access: bool = False

    def get_dashboard_auth_users(self) -> Optional[str]:
        return self.dashboard_auth_users or self.get_secret_from_file("DASHBOARD_AUTH_USERS")

    def get_cr_policy_rules(self) -> List[Any]:
        from phi.k8s.create.rbac_authorization_k8s_io.v1.cluster_role import (
            PolicyRule,
        )

        return [
            PolicyRule(
                api_groups=[""],
                resources=["services", "endpoints", "secrets"],
                verbs=["get", "list", "watch"],
            ),
            PolicyRule(
                api_groups=["extensions", "networking.k8s.io"],
                resources=["ingresses", "ingressclasses"],
                verbs=["get", "list", "watch"],
            ),
            PolicyRule(
                api_groups=["extensions", "networking.k8s.io"],
                resources=["ingresses/status"],
                verbs=["update"],
            ),
            PolicyRule(
                api_groups=["traefik.io", "traefik.containo.us"],
                resources=[
                    "middlewares",
                    "middlewaretcps",
                    "ingressroutes",
                    "traefikservices",
                    "ingressroutetcps",
                    "ingressrouteudps",
                    "tlsoptions",
                    "tlsstores",
                    "serverstransports",
                ],
                verbs=["get", "list", "watch"],
            ),
        ]

    def get_container_args(self) -> Optional[List[str]]:
        if self.command is not None:
            if isinstance(self.command, str):
                return self.command.strip().split(" ")
            return self.command

        container_args = ["--providers.kubernetescrd"]

        if self.access_logs:
            container_args.append("--accesslog")

        if self.http_enabled:
            container_args.append(f"--entrypoints.{self.http_key}.Address=:{self.http_service_port}")
            if self.enable_http_proxy_protocol:
                container_args.append(f"--entrypoints.{self.http_key}.proxyProtocol.insecure=true")
            if self.enable_http_forward_headers:
                container_args.append(f"--entrypoints.{self.http_key}.forwardedHeaders.insecure=true")

        if self.https_enabled:
            container_args.append(f"--entrypoints.{self.https_key}.Address=:{self.https_service_port}")
            if self.enable_https_proxy_protocol:
                container_args.append(f"--entrypoints.{self.https_key}.proxyProtocol.insecure=true")
            if self.enable_https_forward_headers:
                container_args.append(f"--entrypoints.{self.https_key}.forwardedHeaders.insecure=true")
            if self.forward_http_to_https:
                container_args.extend(
                    [
                        f"--entrypoints.{self.http_key}.http.redirections.entryPoint.to={self.https_key}",
                        f"--entrypoints.{self.http_key}.http.redirections.entryPoint.scheme=https",
                    ]
                )

        if self.dashboard_enabled:
            container_args.append("--api=true")
            container_args.append("--api.dashboard=true")
            if self.insecure_api_access:
                container_args.append("--api.insecure")

        return container_args

    def get_secrets(self) -> List[Any]:
        return self.add_secrets or []

    def get_ports(self) -> List[Any]:
        from phi.k8s.create.common.port import CreatePort

        ports: List[CreatePort] = self.add_ports or []

        if self.http_enabled:
            web_port = CreatePort(
                name=self.http_key,
                container_port=self.http_container_port,
                service_port=self.http_service_port,
                target_port=self.http_key,
            )
            if (
                self.service_type in (ServiceType.NODE_PORT, ServiceType.LOAD_BALANCER)
                and self.http_node_port is not None
            ):
                web_port.node_port = self.http_node_port
            ports.append(web_port)

        if self.https_enabled:
            websecure_port = CreatePort(
                name=self.https_key,
                container_port=self.https_container_port,
                service_port=self.https_service_port,
                target_port=self.https_key,
            )
            if (
                self.service_type in (ServiceType.NODE_PORT, ServiceType.LOAD_BALANCER)
                and self.https_node_port is not None
            ):
                websecure_port.node_port = self.https_node_port
            ports.append(websecure_port)

        if self.dashboard_enabled:
            dashboard_port = CreatePort(
                name=self.dashboard_key,
                container_port=self.dashboard_container_port,
                service_port=self.dashboard_service_port,
                target_port=self.dashboard_key,
            )
            if (
                self.service_type in (ServiceType.NODE_PORT, ServiceType.LOAD_BALANCER)
                and self.dashboard_node_port is not None
            ):
                dashboard_port.node_port = self.dashboard_node_port
            ports.append(dashboard_port)

        return ports

    def add_app_resources(self, namespace: str, service_account_name: Optional[str]) -> List[Any]:
        from phi.k8s.create.apiextensions_k8s_io.v1.custom_object import CreateCustomObject

        app_resources = self.add_resources or []

        if self.http_enabled:
            http_ingressroute = CreateCustomObject(
                name=self.http_ingress_name,
                crd=ingressroute_crd,
                spec={
                    "entryPoints": [self.http_key],
                    "routes": self.http_routes,
                },
                app_name=self.get_app_name(),
                namespace=namespace,
            )
            app_resources.append(http_ingressroute)
            logger.debug(f"Added IngressRoute: {http_ingressroute.name}")

        if self.https_enabled:
            https_ingressroute = CreateCustomObject(
                name=self.https_ingress_name,
                crd=ingressroute_crd,
                spec={
                    "entryPoints": [self.https_key],
                    "routes": self.https_routes,
                },
                app_name=self.get_app_name(),
                namespace=namespace,
            )
            app_resources.append(https_ingressroute)
            logger.debug(f"Added IngressRoute: {https_ingressroute.name}")

        if self.add_headers:
            headers_middleware = CreateCustomObject(
                name="header-middleware",
                crd=middleware_crd,
                spec={
                    "headers": self.add_headers,
                },
                app_name=self.get_app_name(),
                namespace=namespace,
            )
            app_resources.append(headers_middleware)
            logger.debug(f"Added Middleware: {headers_middleware.name}")

        if self.dashboard_enabled:
            # create dashboard_auth_middleware if auth provided
            # ref: https://doc.traefik.io/traefik/operations/api/#configuration
            dashboard_auth_middleware = None
            dashboard_auth_users = self.get_dashboard_auth_users()
            if dashboard_auth_users is not None:
                from phi.k8s.create.core.v1.secret import CreateSecret

                dashboard_auth_secret = CreateSecret(
                    secret_name="dashboard-auth-secret",
                    app_name=self.get_app_name(),
                    namespace=namespace,
                    string_data={"users": dashboard_auth_users},
                )
                app_resources.append(dashboard_auth_secret)
                logger.debug(f"Added Secret: {dashboard_auth_secret.secret_name}")

                dashboard_auth_middleware = CreateCustomObject(
                    name="dashboard-auth-middleware",
                    crd=middleware_crd,
                    spec={"basicAuth": {"secret": dashboard_auth_secret.secret_name}},
                    app_name=self.get_app_name(),
                    namespace=namespace,
                )
                app_resources.append(dashboard_auth_middleware)
                logger.debug(f"Added Middleware: {dashboard_auth_middleware.name}")

            dashboard_routes = self.dashboard_routes
            # use default dashboard routes
            if dashboard_routes is None:
                # domain must be provided
                if self.domain_name is not None:
                    dashboard_routes = [
                        {
                            "kind": "Rule",
                            "match": f"Host(`traefik.{self.domain_name}`)",
                            "middlewares": [
                                {
                                    "name": dashboard_auth_middleware.name,
                                    "namespace": namespace,
                                },
                            ]
                            if dashboard_auth_middleware is not None
                            else [],
                            "services": [
                                {
                                    "kind": "TraefikService",
                                    "name": "api@internal",
                                }
                            ],
                        },
                    ]

            dashboard_ingressroute = CreateCustomObject(
                name=self.dashboard_ingress_name,
                crd=ingressroute_crd,
                spec={
                    # "entryPoints": [dashboard_entrypoint],
                    "routes": dashboard_routes,
                },
                app_name=self.get_app_name(),
                namespace=namespace,
            )
            app_resources.append(dashboard_ingressroute)
            logger.debug(f"Added IngressRoute: {dashboard_ingressroute.name}")

        if self.install_crds:
            from phi.k8s.resource.yaml import YamlResource

            if self.yaml_resources is None:
                self.yaml_resources = []
            self.yaml_resources.append(
                YamlResource(
                    name="traefik-crds",
                    url="https://raw.githubusercontent.com/traefik/traefik/v2.10/docs/content/reference/dynamic-configuration/kubernetes-crd-definition-v1.yml",
                )
            )
            logger.debug("Added CRD yaml")

        return app_resources



from typing import Optional, TYPE_CHECKING

if TYPE_CHECKING:
    import kubernetes

from phi.utils.log import logger


class K8sApiClient:
    def __init__(self, context: Optional[str] = None, kubeconfig_path: Optional[str] = None):
        super().__init__()

        try:
            import kubernetes
        except ImportError:
            logger.error(
                "The `kubernetes` package is not installed. "
                "Install using `pip install kubernetes` or `pip install phidata[k8s]`."
            )
            exit(0)

        self.context: Optional[str] = context
        self.kubeconfig_path: Optional[str] = kubeconfig_path
        self.configuration: Optional[kubernetes.client.Configuration] = None

        # kubernetes API clients
        self._api_client: Optional[kubernetes.client.ApiClient] = None
        self._apps_v1_api: Optional[kubernetes.client.AppsV1Api] = None
        self._core_v1_api: Optional[kubernetes.client.CoreV1Api] = None
        self._rbac_auth_v1_api: Optional[kubernetes.client.RbacAuthorizationV1Api] = None
        self._storage_v1_api: Optional[kubernetes.client.StorageV1Api] = None
        self._apiextensions_v1_api: Optional[kubernetes.client.ApiextensionsV1Api] = None
        self._networking_v1_api: Optional[kubernetes.client.NetworkingV1Api] = None
        self._custom_objects_api: Optional[kubernetes.client.CustomObjectsApi] = None
        logger.debug(f"**-+-** K8sApiClient created for {self.context}")

    def create_api_client(self) -> "kubernetes.client.ApiClient":
        """Create a kubernetes.client.ApiClient"""
        logger.debug("Creating kubernetes.client.ApiClient")
        try:
            self.configuration = kubernetes.client.Configuration()
            try:
                kubernetes.config.load_kube_config(
                    config_file=self.kubeconfig_path, client_configuration=self.configuration, context=self.context
                )
            except kubernetes.config.ConfigException:
                # Usually because the context is not in the kubeconfig
                kubernetes.config.load_kube_config(client_configuration=self.configuration)
            logger.debug(f"\thost: {self.configuration.host}")
            self._api_client = kubernetes.client.ApiClient(self.configuration)
            logger.debug(f"\tApiClient: {self._api_client}")
        except Exception as e:
            logger.error(e)

        if self._api_client is None:
            logger.error("Failed to creat Kubernetes ApiClient")
            exit(0)
        return self._api_client

    ######################################################
    # K8s APIs are cached by the class
    ######################################################

    @property
    def api_client(self) -> "kubernetes.client.ApiClient":
        if self._api_client is None:
            self._api_client = self.create_api_client()
        return self._api_client

    @property
    def apps_v1_api(self) -> "kubernetes.client.AppsV1Api":
        if self._apps_v1_api is None:
            self._apps_v1_api = kubernetes.client.AppsV1Api(self.api_client)
        return self._apps_v1_api

    @property
    def core_v1_api(self) -> "kubernetes.client.CoreV1Api":
        if self._core_v1_api is None:
            self._core_v1_api = kubernetes.client.CoreV1Api(self.api_client)
        return self._core_v1_api

    @property
    def rbac_auth_v1_api(self) -> "kubernetes.client.RbacAuthorizationV1Api":
        if self._rbac_auth_v1_api is None:
            self._rbac_auth_v1_api = kubernetes.client.RbacAuthorizationV1Api(self.api_client)
        return self._rbac_auth_v1_api

    @property
    def storage_v1_api(self) -> "kubernetes.client.StorageV1Api":
        if self._storage_v1_api is None:
            self._storage_v1_api = kubernetes.client.StorageV1Api(self.api_client)
        return self._storage_v1_api

    @property
    def apiextensions_v1_api(self) -> "kubernetes.client.ApiextensionsV1Api":
        if self._apiextensions_v1_api is None:
            self._apiextensions_v1_api = kubernetes.client.ApiextensionsV1Api(self.api_client)
        return self._apiextensions_v1_api

    @property
    def networking_v1_api(self) -> "kubernetes.client.NetworkingV1Api":
        if self._networking_v1_api is None:
            self._networking_v1_api = kubernetes.client.NetworkingV1Api(self.api_client)
        return self._networking_v1_api

    @property
    def custom_objects_api(self) -> "kubernetes.client.CustomObjectsApi":
        if self._custom_objects_api is None:
            self._custom_objects_api = kubernetes.client.CustomObjectsApi(self.api_client)
        return self._custom_objects_api



from typing import Any, List

from pydantic import BaseModel

from phi.document.base import Document


class Reader(BaseModel):
    chunk: bool = True
    chunk_size: int = 3000
    separators: List[str] = ["\n", "\n\n", "\r", "\r\n", "\n\r", "\t", " ", "  "]

    def read(self, obj: Any) -> List[Document]:
        raise NotImplementedError

    def clean_text(self, text: str) -> str:
        """Clean the text by replacing multiple newlines with a single newline"""
        import re

        # Replace multiple newlines with a single newline
        cleaned_text = re.sub(r"\n+", "\n", text)
        # Replace multiple spaces with a single space
        cleaned_text = re.sub(r"\s+", " ", cleaned_text)
        # Replace multiple tabs with a single tab
        cleaned_text = re.sub(r"\t+", "\t", cleaned_text)
        # Replace multiple carriage returns with a single carriage return
        cleaned_text = re.sub(r"\r+", "\r", cleaned_text)
        # Replace multiple form feeds with a single form feed
        cleaned_text = re.sub(r"\f+", "\f", cleaned_text)
        # Replace multiple vertical tabs with a single vertical tab
        cleaned_text = re.sub(r"\v+", "\v", cleaned_text)

        return cleaned_text

    def chunk_document(self, document: Document) -> List[Document]:
        """Chunk the document content into smaller documents"""
        content = document.content
        cleaned_content = self.clean_text(content)
        content_length = len(cleaned_content)
        chunked_documents: List[Document] = []
        chunk_number = 1
        chunk_meta_data = document.meta_data

        start = 0
        while start < content_length:
            end = start + self.chunk_size

            # Ensure we're not splitting a word in half
            if end < content_length:
                while end > start and cleaned_content[end] not in [" ", "\n", "\r", "\t"]:
                    end -= 1

            # If the entire chunk is a word, then just split it at self.chunk_size
            if end == start:
                end = start + self.chunk_size

            # If the end is greater than the content length, then set it to the content length
            if end > content_length:
                end = content_length

            chunk = cleaned_content[start:end]
            meta_data = chunk_meta_data.copy()
            meta_data["chunk"] = chunk_number
            meta_data["chunk_size"] = len(chunk)
            chunked_documents.append(
                Document(
                    content=chunk,
                    name=document.name,
                    meta_data=meta_data,
                )
            )
            chunk_number += 1
            start = end
        return chunked_documents



from typing import Optional

from phi.cli.settings import phi_cli_settings
from phi.utils.pickle import pickle_object_to_file, unpickle_object_from_file


class PhiCliCreds:
    def __init__(self, auth_token: str):
        self._auth_token = auth_token

    @property
    def auth_token(self) -> str:
        return self._auth_token


def save_auth_token(auth_token: str):
    # logger.debug(f"Storing {auth_token} to {str(phi_cli_settings.credentials_path)}")
    creds = PhiCliCreds(auth_token)
    pickle_object_to_file(creds, phi_cli_settings.credentials_path)


def read_auth_token() -> Optional[str]:
    # logger.debug(f"Reading token from {str(phi_cli_settings.credentials_path)}")
    creds: PhiCliCreds = unpickle_object_from_file(phi_cli_settings.credentials_path)
    return creds.auth_token



from typing import Any, Dict, List, Optional

from kubernetes.client import CoreV1Api
from kubernetes.client.models.v1_config_map import V1ConfigMap
from kubernetes.client.models.v1_config_map_list import V1ConfigMapList
from kubernetes.client.models.v1_status import V1Status

from phi.k8s.api_client import K8sApiClient
from phi.k8s.resource.base import K8sResource
from phi.utils.log import logger


class ConfigMap(K8sResource):
    """
    ConfigMaps allow you to decouple configuration from image content to keep containerized applications portable.
    In short, they store configs. For config variables which contain sensitive info, use Secrets.

    References:
        * Docs:
            https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#configmap-v1-core
            https://kubernetes.io/docs/tasks/configure-pod-container/configure-pod-configmap/
        * Type: https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_config_map.py
    """

    resource_type: str = "ConfigMap"

    data: Dict[str, Any]

    # List of attributes to include in the K8s manifest
    fields_for_k8s_manifest: List[str] = ["data"]

    def get_k8s_object(self) -> V1ConfigMap:
        """Creates a body for this ConfigMap"""

        # Return a V1ConfigMap object to create a ClusterRole
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_config_map.py
        _v1_config_map = V1ConfigMap(
            api_version=self.api_version.value,
            kind=self.kind.value,
            metadata=self.metadata.get_k8s_object(),
            data=self.data,
        )
        return _v1_config_map

    @staticmethod
    def get_from_cluster(
        k8s_client: K8sApiClient, namespace: Optional[str] = None, **kwargs: str
    ) -> Optional[List[V1ConfigMap]]:
        """Reads ConfigMaps from K8s cluster.

        Args:
            k8s_client: K8sApiClient for the cluster
            namespace: Namespace to use.
        """
        core_v1_api: CoreV1Api = k8s_client.core_v1_api
        # logger.debug(f"core_v1_api: {core_v1_api}")
        cm_list: Optional[V1ConfigMapList] = None
        if namespace:
            # logger.debug(f"Getting CMs for ns: {namespace}")
            cm_list = core_v1_api.list_namespaced_config_map(namespace=namespace)
        else:
            # logger.debug("Getting CMs for all namespaces")
            cm_list = core_v1_api.list_config_map_for_all_namespaces()

        config_maps: Optional[List[V1ConfigMap]] = None
        if cm_list:
            config_maps = cm_list.items
        # logger.debug(f"config_maps: {config_maps}")
        # logger.debug(f"config_maps type: {type(config_maps)}")
        return config_maps

    def _create(self, k8s_client: K8sApiClient) -> bool:
        core_v1_api: CoreV1Api = k8s_client.core_v1_api
        # logger.debug(f"core_v1_api: {core_v1_api}")
        k8s_object: V1ConfigMap = self.get_k8s_object()
        # logger.debug(f"k8s_object: {k8s_object}")
        namespace = self.get_namespace()
        # logger.debug(f"namespace: {namespace}")

        logger.debug("Creating: {}".format(self.get_resource_name()))
        v1_config_map: V1ConfigMap = core_v1_api.create_namespaced_config_map(
            namespace=namespace,
            body=k8s_object,
            async_req=self.async_req,
            pretty=self.pretty,
        )
        # logger.debug("Created: {}".format(v1_config_map))
        if v1_config_map.metadata.creation_timestamp is not None:
            logger.debug("ConfigMap Created")
            self.active_resource = v1_config_map
            return True
        logger.error("ConfigMap could not be created")
        return False

    def _read(self, k8s_client: K8sApiClient) -> Optional[V1ConfigMap]:
        """Returns the "Active" ConfigMap from the cluster"""

        namespace = self.get_namespace()
        active_resource: Optional[V1ConfigMap] = None
        active_resources: Optional[List[V1ConfigMap]] = self.get_from_cluster(
            k8s_client=k8s_client,
            namespace=namespace,
        )
        # logger.debug(f"active_resources: {active_resources}")
        if active_resources is None:
            return None

        active_resources_dict = {_cm.metadata.name: _cm for _cm in active_resources}

        cm_name = self.get_resource_name()
        if cm_name in active_resources_dict:
            active_resource = active_resources_dict[cm_name]
            self.active_resource = active_resource
            logger.debug(f"Found active {cm_name}")
        return active_resource

    def _update(self, k8s_client: K8sApiClient) -> bool:
        core_v1_api: CoreV1Api = k8s_client.core_v1_api
        cm_name = self.get_resource_name()
        k8s_object: V1ConfigMap = self.get_k8s_object()
        namespace = self.get_namespace()

        logger.debug("Updating: {}".format(cm_name))
        v1_config_map: V1ConfigMap = core_v1_api.patch_namespaced_config_map(
            name=cm_name,
            namespace=namespace,
            body=k8s_object,
            async_req=self.async_req,
            pretty=self.pretty,
        )
        # logger.debug("Updated:\n{}".format(pformat(v1_config_map.to_dict(), indent=2)))
        if v1_config_map.metadata.creation_timestamp is not None:
            logger.debug("ConfigMap Updated")
            self.active_resource = v1_config_map
            return True
        logger.error("ConfigMap could not be updated")
        return False

    def _delete(self, k8s_client: K8sApiClient) -> bool:
        core_v1_api: CoreV1Api = k8s_client.core_v1_api
        cm_name = self.get_resource_name()
        namespace = self.get_namespace()

        logger.debug("Deleting: {}".format(cm_name))
        self.active_resource = None
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_status.py
        delete_status: V1Status = core_v1_api.delete_namespaced_config_map(
            name=cm_name,
            namespace=namespace,
            async_req=self.async_req,
            pretty=self.pretty,
        )
        logger.debug("delete_status: {}".format(delete_status.status))
        if delete_status.status == "Success":
            logger.debug("ConfigMap Deleted")
            return True
        logger.error("ConfigMap could not be deleted")
        return False



from pathlib import Path
from typing import List, Optional, Any, Dict

from pydantic import BaseModel, Field, ConfigDict

from phi.utils.log import logger


class KubeconfigClusterConfig(BaseModel):
    server: str
    certificate_authority_data: str = Field(..., alias="certificate-authority-data")

    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)


class KubeconfigCluster(BaseModel):
    name: str
    cluster: KubeconfigClusterConfig


class KubeconfigUser(BaseModel):
    name: str
    user: Dict[str, Any]


class KubeconfigContextSpec(BaseModel):
    """Each Kubeconfig context is made of (cluster, user, namespace).
    It should be read as:
        Use the credentials of the "user"
        to access the "namespace"
        of the "cluster"
    """

    cluster: Optional[str]
    user: Optional[str]
    namespace: Optional[str] = None


class KubeconfigContext(BaseModel):
    """A context element in a kubeconfig file is used to group access parameters under a
    convenient name. Each context has three parameters: cluster, namespace, and user.
    By default, the kubectl command-line tool uses parameters from the current context
    to communicate with the cluster.
    """

    name: str
    context: KubeconfigContextSpec


class Kubeconfig(BaseModel):
    """
    We configure access to K8s clusters using a Kubeconfig.
    This configuration can be stored in a file or an object.
    A Kubeconfig stores information about clusters, users, namespaces, and authentication mechanisms,

    Locally the kubeconfig file is usually stored at ~/.kube/config
    View your local kubeconfig using `kubectl config view`

    References:
        * Docs:
            https://kubernetes.io/docs/tasks/access-application-cluster/configure-access-multiple-clusters/
        * Go Doc: https://godoc.org/k8s.io/client-go/tools/clientcmd/api#Config
    """

    api_version: str = Field("v1", alias="apiVersion")
    kind: str = "Config"
    clusters: List[KubeconfigCluster] = []
    users: List[KubeconfigUser] = []
    contexts: List[KubeconfigContext] = []
    current_context: Optional[str] = Field(None, alias="current-context")
    preferences: dict = {}

    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    @classmethod
    def read_from_file(cls: Any, file_path: Path, create_if_not_exists: bool = True) -> Optional[Any]:
        if file_path is not None:
            if not file_path.exists():
                if create_if_not_exists:
                    logger.info(f"Creating: {file_path}")
                    file_path.parent.mkdir(parents=True, exist_ok=True)
                    file_path.touch()
                else:
                    logger.error(f"File does not exist: {file_path}")
                    return None
            if file_path.exists() and file_path.is_file():
                try:
                    import yaml

                    logger.info(f"Reading {file_path}")
                    kubeconfig_dict = yaml.safe_load(file_path.read_text())
                    if kubeconfig_dict is not None and isinstance(kubeconfig_dict, dict):
                        kubeconfig = cls(**kubeconfig_dict)
                        return kubeconfig
                except Exception as e:
                    logger.error(f"Error reading {file_path}")
                    logger.error(e)
        else:
            logger.warning(f"Kubeconfig invalid: {file_path}")
        return None

    def write_to_file(self, file_path: Path) -> bool:
        """Writes the kubeconfig to file_path"""
        if file_path is not None:
            try:
                import yaml

                kubeconfig_dict = self.model_dump(exclude_none=True, by_alias=True)
                file_path.parent.mkdir(parents=True, exist_ok=True)
                file_path.write_text(yaml.safe_dump(kubeconfig_dict))
                logger.info(f"Updated: {file_path}")
                return True
            except Exception as e:
                logger.error(f"Error writing {file_path}")
                logger.error(e)
        else:
            logger.error(f"Kubeconfig invalid: {file_path}")
        return False



from pathlib import Path
from typing import Optional, Any

from phi.k8s.api_client import K8sApiClient
from phi.k8s.enums.api_version import ApiVersion
from phi.k8s.enums.kind import Kind
from phi.k8s.resource.base import K8sResource
from phi.k8s.resource.meta.v1.object_meta import ObjectMeta


class YamlResource(K8sResource):
    resource_type: str = "Yaml"

    api_version: ApiVersion = ApiVersion.NA
    kind: Kind = Kind.YAML
    metadata: ObjectMeta = ObjectMeta()

    file: Optional[Path] = None
    dir: Optional[Path] = None
    url: Optional[str] = None

    @staticmethod
    def get_from_cluster(k8s_client: K8sApiClient, namespace: Optional[str] = None, **kwargs) -> None:
        # Not implemented for YamlResources
        return None

    def _create(self, k8s_client: K8sApiClient) -> bool:
        return True

    def _read(self, k8s_client: K8sApiClient) -> Optional[Any]:
        return None

    def _update(self, k8s_client: K8sApiClient) -> bool:
        return True

    def _delete(self, k8s_client: K8sApiClient) -> bool:
        return True



from typing import List

from phi.llm.function.registry import FunctionRegistry
from phi.utils.log import logger


class ShellScriptsRegistry(FunctionRegistry):
    def __init__(self):
        super().__init__(name="shell_script_registry")
        self.register(self.run_shell_command)

    def run_shell_command(self, args: List[str], tail: int = 100) -> str:
        """Runs a shell command and returns the output or error.

        :param args: The command to run as a list of strings.
        :param tail: The number of lines to return from the output.
        :return: The output of the command.
        """
        logger.info(f"Running shell command: {args}")

        import subprocess

        try:
            result = subprocess.run(args, capture_output=True, text=True)
            logger.debug(f"Result: {result}")
            logger.debug(f"Return code: {result.returncode}")
            if result.returncode != 0:
                return f"Error: {result.stderr}"

            # return only the last n lines of the output
            return "\n".join(result.stdout.split("\n")[-tail:])
        except Exception as e:
            logger.warning(f"Failed to run shell command: {e}")
            return f"Error: {e}"



from typing import Optional, List, Union
from hashlib import md5

try:
    from sqlalchemy.dialects import postgresql
    from sqlalchemy.engine import create_engine, Engine
    from sqlalchemy.inspection import inspect
    from sqlalchemy.orm import Session, sessionmaker
    from sqlalchemy.schema import MetaData, Table, Column
    from sqlalchemy.sql.expression import text, func, select
    from sqlalchemy.types import DateTime, String
except ImportError:
    raise ImportError("`sqlalchemy` not installed")

try:
    from pgvector.sqlalchemy import Vector
except ImportError:
    raise ImportError("`pgvector` not installed")

from phi.document import Document
from phi.embedder import Embedder
from phi.embedder.openai import OpenAIEmbedder
from phi.vectordb.base import VectorDb
from phi.vectordb.distance import Distance
from phi.vectordb.pgvector.index import Ivfflat, HNSW
from phi.utils.log import logger


class PgVector(VectorDb):
    def __init__(
        self,
        collection: str,
        schema: Optional[str] = "llm",
        db_url: Optional[str] = None,
        db_engine: Optional[Engine] = None,
        embedder: Embedder = OpenAIEmbedder(),
        distance: Distance = Distance.cosine,
        index: Optional[Union[Ivfflat, HNSW]] = Ivfflat(),
    ):
        _engine: Optional[Engine] = db_engine
        if _engine is None and db_url is not None:
            _engine = create_engine(db_url)

        if _engine is None:
            raise ValueError("Must provide either db_url or db_engine")

        # Collection attributes
        self.collection: str = collection
        self.schema: Optional[str] = schema

        # Database attributes
        self.db_url: Optional[str] = db_url
        self.db_engine: Engine = _engine
        self.metadata: MetaData = MetaData(schema=self.schema)

        # Embedder for embedding the document contents
        self.embedder: Embedder = embedder
        self.dimensions: int = self.embedder.dimensions

        # Distance metric
        self.distance: Distance = distance

        # Index for the collection
        self.index: Optional[Union[Ivfflat, HNSW]] = index

        # Database session
        self.Session: sessionmaker[Session] = sessionmaker(bind=self.db_engine)

        # Database table for the collection
        self.table: Table = self.get_table()

    def get_table(self) -> Table:
        return Table(
            self.collection,
            self.metadata,
            Column("name", String),
            Column("meta_data", postgresql.JSONB, server_default=text("'{}'::jsonb")),
            Column("content", postgresql.TEXT),
            Column("embedding", Vector(self.dimensions)),
            Column("usage", postgresql.JSONB),
            Column("created_at", DateTime(timezone=True), server_default=text("now()")),
            Column("updated_at", DateTime(timezone=True), onupdate=text("now()")),
            Column("content_hash", String),
            extend_existing=True,
        )

    def table_exists(self) -> bool:
        logger.debug(f"Checking if table exists: {self.table.name}")
        try:
            return inspect(self.db_engine).has_table(self.table.name, schema=self.schema)
        except Exception as e:
            logger.error(e)
            return False

    def create(self) -> None:
        if not self.table_exists():
            with self.Session() as sess:
                with sess.begin():
                    logger.debug("Creating extension: vector")
                    sess.execute(text("create extension if not exists vector;"))
                    if self.schema is not None:
                        logger.debug(f"Creating schema: {self.schema}")
                        sess.execute(text(f"create schema if not exists {self.schema};"))
            logger.debug(f"Creating table: {self.collection}")
            self.table.create(self.db_engine)

    def doc_exists(self, document: Document) -> bool:
        """
        Validating if the document exists or not

        Args:
            document (Document): Document to validate
        """
        columns = [self.table.c.name, self.table.c.content_hash]
        with self.Session() as sess:
            with sess.begin():
                cleaned_content = document.content.replace("\x00", "\uFFFD")
                stmt = select(*columns).where(self.table.c.content_hash == md5(cleaned_content.encode()).hexdigest())
                result = sess.execute(stmt).first()
                return result is not None

    def name_exists(self, name: str) -> bool:
        """
        Validate if a row with this name exists or not

        Args:
            name (str): Name to validate
        """
        columns = [self.table.c.name]
        with self.Session() as sess:
            with sess.begin():
                stmt = select(*columns).where(self.table.c.name == name)
                result = sess.execute(stmt).first()
                return result is not None

    def insert(self, documents: List[Document], batch_size: int = 10) -> None:
        with self.Session() as sess:
            counter = 0
            for document in documents:
                document.embed(embedder=self.embedder)
                cleaned_content = document.content.replace("\x00", "\uFFFD")
                stmt = postgresql.insert(self.table).values(
                    name=document.name,
                    meta_data=document.meta_data,
                    content=cleaned_content,
                    embedding=document.embedding,
                    usage=document.usage,
                    content_hash=md5(cleaned_content.encode()).hexdigest(),
                )
                sess.execute(stmt)
                counter += 1
                logger.debug(f"Inserted document: {document.name} ({document.meta_data})")

                # Commit every `batch_size` documents
                if counter >= batch_size:
                    sess.commit()
                    logger.debug(f"Committed {counter} documents")
                    counter = 0

            # Commit any remaining documents
            if counter > 0:
                sess.commit()
                logger.debug(f"Committed {counter} documents")

    def upsert(self, documents: List[Document]) -> None:
        """
        Upsert documents into the database.

        Args:
            documents (List[Document]): List of documents to upsert
        """
        with self.Session() as sess:
            with sess.begin():
                for document in documents:
                    document.embed(embedder=self.embedder)
                    cleaned_content = document.content.replace("\x00", "\uFFFD")
                    stmt = postgresql.insert(self.table).values(
                        name=document.name,
                        meta_data=document.meta_data,
                        content=cleaned_content,
                        embedding=document.embedding,
                        usage=document.usage,
                        content_hash=md5(cleaned_content.encode()).hexdigest(),
                    )
                    stmt = stmt.on_conflict_do_update(
                        index_elements=["name", "content_hash"],
                        set_=dict(
                            meta_data=document.meta_data,
                            content=stmt.excluded.content,
                            embedding=stmt.excluded.embedding,
                            usage=stmt.excluded.usage,
                        ),
                    )
                    sess.execute(stmt)
                    logger.debug(f"Upserted document: {document.name} ({document.meta_data})")

    def search(self, query: str, limit: int = 5) -> List[Document]:
        query_embedding = self.embedder.get_embedding(query)
        if query_embedding is None:
            logger.error(f"Error getting embedding for Query: {query}")
            return []

        columns = [
            self.table.c.name,
            self.table.c.meta_data,
            self.table.c.content,
            self.table.c.embedding,
            self.table.c.usage,
        ]

        stmt = select(*columns)
        if self.distance == Distance.l2:
            stmt = stmt.order_by(self.table.c.embedding.max_inner_product(query_embedding))
        if self.distance == Distance.cosine:
            stmt = stmt.order_by(self.table.c.embedding.cosine_distance(query_embedding))
        if self.distance == Distance.max_inner_product:
            stmt = stmt.order_by(self.table.c.embedding.max_inner_product(query_embedding))

        stmt = stmt.limit(limit=limit)
        logger.debug(f"Query: {stmt}")

        # Get neighbors
        with self.Session() as sess:
            with sess.begin():
                if self.index is not None and isinstance(self.index, Ivfflat):
                    sess.execute(text(f"SET LOCAL ivfflat.probes = {self.index.probes}"))
                neighbors = sess.execute(stmt).fetchall() or []

        # Build search results
        search_results: List[Document] = []
        for neighbor in neighbors:
            search_results.append(
                Document(
                    name=neighbor.name,
                    meta_data=neighbor.meta_data,
                    content=neighbor.content,
                    embedder=self.embedder,
                    embedding=neighbor.embedding,
                    usage=neighbor.usage,
                )
            )

        return search_results

    def delete(self) -> None:
        if self.table_exists():
            logger.debug(f"Deleting table: {self.collection}")
            self.table.drop(self.db_engine)

    def exists(self) -> bool:
        return self.table_exists()

    def get_count(self) -> int:
        with self.Session() as sess:
            with sess.begin():
                stmt = select(func.count(self.table.c.name)).select_from(self.table)
                result = sess.execute(stmt).scalar()
                if result is not None:
                    return int(result)
                return 0

    def optimize(self) -> None:
        from math import sqrt

        if self.index is None:
            return

        if isinstance(self.index, Ivfflat):
            num_lists = self.index.lists
            if self.index.dynamic_lists:
                total_records = self.get_count()
                logger.debug(f"Number of records: {total_records}")
                if total_records < 1000000:
                    num_lists = int(total_records / 1000)
                elif total_records > 1000000:
                    num_lists = int(sqrt(total_records))

            index_distance = "vector_cosine_ops"
            if self.distance == Distance.l2:
                index_distance = "vector_l2_ops"
            if self.distance == Distance.max_inner_product:
                index_distance = "vector_ip_ops"
            with self.Session() as sess:
                with sess.begin():
                    logger.debug(f"Setting configuration: {self.index.configuration}")
                    for key, value in self.index.configuration.items():
                        sess.execute(text(f"SET {key} = '{value}';"))
                    logger.debug(
                        f"Creating Index with lists: {num_lists}, probes: {self.index.probes} "
                        f"and distance metric: {index_distance}"
                    )
                    sess.execute(text(f"SET ivfflat.probes = {self.index.probes};"))
                    sess.execute(
                        text(
                            f"CREATE INDEX ON {self.table} USING ivfflat (embedding \
                                {index_distance}) WITH (lists = {num_lists});"
                        )
                    )
        logger.debug("Optimized Vector DB")



from pathlib import Path
from typing import List

from phi.document.base import Document
from phi.document.reader.base import Reader
from phi.utils.log import logger


class PDFReader(Reader):
    """Reader for PDF files"""

    def read(self, path: Path) -> List[Document]:
        if not path:
            raise ValueError("No path provided")

        if not path.exists():
            raise FileNotFoundError(f"Could not find file: {path}")

        try:
            from pypdf import PdfReader as DocumentReader  # noqa: F401
        except ImportError:
            raise ImportError("`pypdf` not installed")

        logger.info(f"Reading: {path}")
        doc_name = path.name.split(".")[0]
        doc_reader = DocumentReader(path)

        documents = [
            Document(
                name=doc_name,
                meta_data={"page": page_number},
                content=page.extract_text(),
            )
            for page_number, page in enumerate(doc_reader.pages, start=1)
        ]
        if self.chunk:
            chunked_documents = []
            for document in documents:
                chunked_documents.extend(self.chunk_document(document))
            return chunked_documents
        return documents



import json
from typing import List, Optional

from phi.document import Document
from phi.knowledge.arxiv import ArxivKnowledgeBase
from phi.llm.function.registry import FunctionRegistry
from phi.utils.log import logger


class ArxivRegistry(FunctionRegistry):
    def __init__(self, knowledge_base: Optional[ArxivKnowledgeBase] = None):
        super().__init__(name="arxiv_registry")
        self.knowledge_base: Optional[ArxivKnowledgeBase] = knowledge_base

        if self.knowledge_base is not None and isinstance(self.knowledge_base, ArxivKnowledgeBase):
            self.register(self.search_arxiv_and_update_knowledge_base)
        else:
            self.register(self.search_arxiv)

    def search_arxiv_and_update_knowledge_base(self, topic: str) -> str:
        """This function searches arXiv for a topic, adds the results to the knowledge base and returns them.

        USE THIS FUNCTION TO GET INFORMATION WHICH DOES NOT EXIST.

        :param topic: The topic to search arXiv and add to knowledge base.
        :return: Relevant documents from arXiv knowledge base.
        """
        if self.knowledge_base is None:
            return "Knowledge base not provided"

        logger.debug(f"Adding to knowledge base: {topic}")
        self.knowledge_base.queries.append(topic)
        logger.debug("Loading knowledge base.")
        self.knowledge_base.load(recreate=False)
        logger.debug(f"Searching knowledge base: {topic}")
        relevant_docs: List[Document] = self.knowledge_base.search(query=topic)
        return json.dumps([doc.to_dict() for doc in relevant_docs])

    def search_arxiv(self, query: str, max_results: int = 5) -> str:
        """
        Searches arXiv for a query.

        :param query: The query to search for.
        :param max_results: The maximum number of results to return.
        :return: Relevant documents from arXiv.
        """
        from phi.document.reader.arxiv import ArxivReader

        arxiv = ArxivReader(max_results=max_results)

        logger.debug(f"Searching arxiv for: {query}")
        relevant_docs: List[Document] = arxiv.read(query=query)
        return json.dumps([doc.to_dict() for doc in relevant_docs])



from pathlib import Path
from typing import Optional, List, Dict

from pydantic import field_validator, Extra
from pydantic_settings import BaseSettings, SettingsConfigDict
from pydantic_core.core_schema import FieldValidationInfo

from phi.api.schemas.workspace import WorkspaceSchema


class WorkspaceSettings(BaseSettings):
    """
    -*- Workspace settings
    Initialize workspace settings by:
    1. Creating a WorkspaceSettings object
    2. Using Environment variables
    3. Using the .env file
    """

    # Workspace name: used for naming cloud resources
    ws_name: str
    # Path to the workspace root
    ws_root: Path
    # Workspace git repo url: used to git-sync DAGs and Charts
    ws_repo: Optional[str] = None
    # Path to important directories relative to the ws_root
    scripts_dir: str = "scripts"
    storage_dir: str = "storage"
    workflows_dir: str = "workflows"
    workspace_dir: str = "workspace"
    # default env for phi ws commands
    default_env: Optional[str] = "dev"
    # default infra for phi ws commands
    default_infra: Optional[str] = None
    #
    # -*- Dev settings
    #
    dev_env: str = "dev"
    # Dev git repo branch: used to git-sync DAGs and Charts
    dev_branch: str = "main"
    # Key for naming dev resources
    dev_key: Optional[str] = None
    # Tags for dev resources
    dev_tags: Optional[Dict[str, str]] = None
    # Domain for the dev platform
    dev_domain: Optional[str] = None
    #
    # -*- Dev Apps
    #
    dev_airflow_enabled: bool = False
    dev_api_enabled: bool = False
    dev_app_enabled: bool = False
    dev_db_enabled: bool = False
    dev_jupyter_enabled: bool = False
    dev_redis_enabled: bool = False
    dev_superset_enabled: bool = False
    dev_traefik_enabled: bool = False
    dev_qdrant_enabled: bool = False
    #
    # -*- Staging settings
    #
    stg_env: str = "stg"
    # Staging git repo branch: used to git-sync DAGs and Charts
    stg_branch: str = "main"
    # Key for naming staging resources
    stg_key: Optional[str] = None
    # Tags for staging resources
    stg_tags: Optional[Dict[str, str]] = None
    # Domain for the staging platform
    stg_domain: Optional[str] = None
    #
    # -*- Staging Apps
    #
    stg_airflow_enabled: bool = False
    stg_api_enabled: bool = False
    stg_app_enabled: bool = False
    stg_db_enabled: bool = False
    stg_jupyter_enabled: bool = False
    stg_redis_enabled: bool = False
    stg_superset_enabled: bool = False
    stg_traefik_enabled: bool = False
    stg_whoami_enabled: bool = False
    #
    # -*- Production settings
    #
    prd_env: str = "prd"
    # Production git repo branch: used to git-sync DAGs and Charts
    prd_branch: str = "main"
    # Key for naming production resources
    prd_key: Optional[str] = None
    # Tags for production resources
    prd_tags: Optional[Dict[str, str]] = None
    # Domain for the production platform
    prd_domain: Optional[str] = None
    #
    # -*- Production Apps
    #
    prd_airflow_enabled: bool = False
    prd_api_enabled: bool = False
    prd_app_enabled: bool = False
    prd_db_enabled: bool = False
    prd_jupyter_enabled: bool = False
    prd_redis_enabled: bool = False
    prd_superset_enabled: bool = False
    prd_traefik_enabled: bool = False
    prd_whoami_enabled: bool = False
    #
    # -*- AWS settings
    #
    # Region for AWS resources
    aws_region: Optional[str] = None
    # Availability Zones for AWS resources
    aws_az1: Optional[str] = None
    aws_az2: Optional[str] = None
    aws_az3: Optional[str] = None
    aws_az4: Optional[str] = None
    aws_az5: Optional[str] = None
    # Public subnets. 1 in each AZ.
    public_subnets: List[str] = []
    # Private subnets. 1 in each AZ.
    private_subnets: List[str] = []
    # Subnet IDs. 1 in each AZ.
    # Derived from public and private subnets if not provided.
    subnet_ids: Optional[List[str]] = None
    # Security Groups
    security_groups: Optional[List[str]] = None
    aws_profile: Optional[str] = None
    aws_config_file: Optional[str] = None
    aws_shared_credentials_file: Optional[str] = None
    # -*- Cli settings
    # Set to True if `phi` should continue creating
    # resources after a resource creation has failed
    continue_on_create_failure: bool = False
    # Set to True if `phi` should continue deleting
    # resources after a resource deleting has failed
    # Defaults to True because we normally want to continue deleting
    continue_on_delete_failure: bool = True
    # Set to True if `phi` should continue patching
    # resources after a resource patch has failed
    continue_on_patch_failure: bool = False
    #
    # -*- Image Settings
    #
    # Repository for images
    image_repo: str = "phidata"
    # Suffix added to the image name
    image_suffix: Optional[str] = None
    # Build images locally
    build_images: bool = False
    # Push images after building
    push_images: bool = False
    # Skip cache when building images
    skip_image_cache: bool = False
    # Force pull images in FROM
    force_pull_images: bool = False
    #
    # -*- Other Settings
    #
    use_cache: bool = True
    # WorkspaceSchema provided by the api
    ws_schema: Optional[WorkspaceSchema] = None

    model_config = SettingsConfigDict(extra=Extra.allow)

    @field_validator("dev_key", mode="before")
    def set_dev_key(cls, dev_key, info: FieldValidationInfo):
        if dev_key is not None:
            return dev_key

        ws_name = info.data.get("ws_name")
        if ws_name is None:
            raise ValueError("ws_name invalid")

        dev_env = info.data.get("dev_env")
        if dev_env is None:
            raise ValueError("dev_env invalid")

        return f"{ws_name}-{dev_env}"

    @field_validator("dev_tags", mode="before")
    def set_dev_tags(cls, dev_tags, info: FieldValidationInfo):
        if dev_tags is not None:
            return dev_tags

        ws_name = info.data.get("ws_name")
        if ws_name is None:
            raise ValueError("ws_name invalid")

        dev_env = info.data.get("dev_env")
        if dev_env is None:
            raise ValueError("dev_env invalid")

        return {
            "Env": dev_env,
            "Project": ws_name,
        }

    @field_validator("stg_key", mode="before")
    def set_stg_key(cls, stg_key, info: FieldValidationInfo):
        if stg_key is not None:
            return stg_key

        ws_name = info.data.get("ws_name")
        if ws_name is None:
            raise ValueError("ws_name invalid")

        stg_env = info.data.get("stg_env")
        if stg_env is None:
            raise ValueError("stg_env invalid")

        return f"{ws_name}-{stg_env}"

    @field_validator("stg_tags", mode="before")
    def set_stg_tags(cls, stg_tags, info: FieldValidationInfo):
        if stg_tags is not None:
            return stg_tags

        ws_name = info.data.get("ws_name")
        if ws_name is None:
            raise ValueError("ws_name invalid")

        stg_env = info.data.get("stg_env")
        if stg_env is None:
            raise ValueError("stg_env invalid")

        return {
            "Env": stg_env,
            "Project": ws_name,
        }

    @field_validator("prd_key", mode="before")
    def set_prd_key(cls, prd_key, info: FieldValidationInfo):
        if prd_key is not None:
            return prd_key

        ws_name = info.data.get("ws_name")
        if ws_name is None:
            raise ValueError("ws_name invalid")

        prd_env = info.data.get("prd_env")
        if prd_env is None:
            raise ValueError("prd_env invalid")

        return f"{ws_name}-{prd_env}"

    @field_validator("prd_tags", mode="before")
    def set_prd_tags(cls, prd_tags, info: FieldValidationInfo):
        if prd_tags is not None:
            return prd_tags

        ws_name = info.data.get("ws_name")
        if ws_name is None:
            raise ValueError("ws_name invalid")

        prd_env = info.data.get("prd_env")
        if prd_env is None:
            raise ValueError("prd_env invalid")

        return {
            "Env": prd_env,
            "Project": ws_name,
        }

    @field_validator("subnet_ids", mode="before")
    def set_subnet_ids(cls, subnet_ids, info: FieldValidationInfo):
        if subnet_ids is not None:
            return subnet_ids

        public_subnets = info.data.get("public_subnets", [])
        private_subnets = info.data.get("private_subnets", [])

        return public_subnets + private_subnets



from typing import List, Optional, Union, Any

from kubernetes.client import NetworkingV1Api
from kubernetes.client.models.v1_ingress import V1Ingress
from kubernetes.client.models.v1_ingress_backend import V1IngressBackend
from kubernetes.client.models.v1_ingress_list import V1IngressList
from kubernetes.client.models.v1_ingress_rule import V1IngressRule
from kubernetes.client.models.v1_ingress_spec import V1IngressSpec
from kubernetes.client.models.v1_ingress_tls import V1IngressTLS
from kubernetes.client.models.v1_status import V1Status

from phi.k8s.api_client import K8sApiClient
from phi.k8s.resource.base import K8sResource, K8sObject
from phi.utils.log import logger


class ServiceBackendPort(K8sObject):
    """
    Reference:
    - https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#servicebackendport-v1-networking-k8s-io
    """

    resource_type: str = "ServiceBackendPort"

    number: int
    name: Optional[str] = None


class IngressServiceBackend(K8sObject):
    """
    Reference:
    - https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#ingressbackend-v1-networking-k8s-io
    """

    resource_type: str = "IngressServiceBackend"

    service_name: str
    service_port: Union[int, str]


class IngressBackend(K8sObject):
    """
    Reference:
    - https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#ingressbackend-v1-networking-k8s-io
    """

    resource_type: str = "IngressBackend"

    service: Optional[V1IngressBackend] = None
    resource: Optional[Any] = None


class HTTPIngressPath(K8sObject):
    """
    Reference:
    - https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#httpingresspath-v1-networking-k8s-io
    """

    resource_type: str = "HTTPIngressPath"

    path: Optional[str] = None
    path_type: Optional[str] = None
    backend: IngressBackend


class HTTPIngressRuleValue(K8sObject):
    """
    Reference:
    - https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#httpingressrulevalue-v1-networking-k8s-io
    """

    resource_type: str = "HTTPIngressRuleValue"

    paths: List[HTTPIngressPath]


class IngressRule(K8sObject):
    """
    Reference:
    - https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#ingressrule-v1-networking-k8s-io
    """

    resource_type: str = "IngressRule"

    host: Optional[str] = None
    http: Optional[HTTPIngressRuleValue] = None


class IngressSpec(K8sObject):
    """
    Reference:
    - https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#ingressspec-v1-core
    """

    resource_type: str = "IngressSpec"

    # DefaultBackend is the backend that should handle requests that don't match any rule.
    # If Rules are not specified, DefaultBackend must be specified.
    # If DefaultBackend is not set, the handling of requests that do not match any of
    # the rules will be up to the Ingress controller.
    default_backend: Optional[V1IngressBackend] = None
    # IngressClassName is the name of the IngressClass cluster resource.
    # The associated IngressClass defines which controller will implement the resource.
    # This replaces the deprecated `kubernetes.io/ingress.class` annotation.
    # For backwards compatibility, when that annotation is set, it must be given precedence over this field.
    ingress_class_name: Optional[str] = None
    # A list of host rules used to configure the Ingress. If unspecified, or no rule matches,
    # all traffic is sent to the default backend.
    rules: Optional[List[V1IngressRule]] = None
    # TLS configuration. The Ingress only supports a single TLS port, 443.
    # If multiple members of this list specify different hosts, they will be multiplexed on the
    # same port according to the hostname specified through the SNI TLS extension, if the ingress controller
    # fulfilling the ingress supports SNI.
    tls: Optional[List[V1IngressTLS]] = None

    def get_k8s_object(self) -> V1IngressSpec:
        # Return a V1IngressSpec object
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_ingress_spec.py

        _v1_ingress_spec = V1IngressSpec(
            default_backend=self.default_backend,
            ingress_class_name=self.ingress_class_name,
            rules=self.rules,
            tls=self.tls,
        )
        return _v1_ingress_spec


class Ingress(K8sResource):
    """
    References:
    - Docs: https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#ingress-v1-networking-k8s-io
    - Type: https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_ingress.py
    """

    resource_type: str = "Ingress"

    spec: IngressSpec

    # List of attributes to include in the K8s manifest
    fields_for_k8s_manifest: List[str] = ["spec"]

    def get_k8s_object(self) -> V1Ingress:
        """Creates a body for this Ingress"""

        # Return a V1Ingress object to create a ClusterRole
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_ingress.py
        _v1_ingress = V1Ingress(
            api_version=self.api_version.value,
            kind=self.kind.value,
            metadata=self.metadata.get_k8s_object(),
            spec=self.spec.get_k8s_object(),
        )
        return _v1_ingress

    @staticmethod
    def get_from_cluster(
        k8s_client: K8sApiClient, namespace: Optional[str] = None, **kwargs
    ) -> Optional[List[V1Ingress]]:
        """Reads Ingress from K8s cluster.

        Args:
            k8s_client: K8sApiClient for the cluster
            namespace: Namespace to use.
        """
        networking_v1_api: NetworkingV1Api = k8s_client.networking_v1_api
        ingress_list: Optional[V1IngressList] = None
        if namespace:
            logger.debug(f"Getting ingress for ns: {namespace}")
            ingress_list = networking_v1_api.list_namespaced_ingress(namespace=namespace)
        else:
            logger.debug("Getting ingress for all namespaces")
            ingress_list = networking_v1_api.list_ingress_for_all_namespaces()

        ingress: Optional[List[V1Ingress]] = None
        if ingress_list:
            ingress = ingress_list.items
        logger.debug(f"ingress: {ingress}")
        logger.debug(f"ingress type: {type(ingress)}")
        return ingress

    def _create(self, k8s_client: K8sApiClient) -> bool:
        networking_v1_api: NetworkingV1Api = k8s_client.networking_v1_api
        k8s_object: V1Ingress = self.get_k8s_object()
        namespace = self.get_namespace()

        logger.debug("Creating: {}".format(self.get_resource_name()))
        v1_ingress: V1Ingress = networking_v1_api.create_namespaced_ingress(
            namespace=namespace,
            body=k8s_object,
            async_req=self.async_req,
            pretty=self.pretty,
        )
        logger.debug("Created: {}".format(v1_ingress))
        if v1_ingress.metadata.creation_timestamp is not None:
            logger.debug("Ingress Created")
            self.active_resource = v1_ingress
            return True
        logger.error("Ingress could not be created")
        return False

    def _read(self, k8s_client: K8sApiClient) -> Optional[V1Ingress]:
        """Returns the "Active" Ingress from the cluster"""

        namespace = self.get_namespace()
        active_resource: Optional[V1Ingress] = None
        active_resources: Optional[List[V1Ingress]] = self.get_from_cluster(
            k8s_client=k8s_client,
            namespace=namespace,
        )
        logger.debug(f"Active Resources: {active_resources}")
        if active_resources is None:
            return None

        active_resources_dict = {_ingress.metadata.name: _ingress for _ingress in active_resources}

        ingress_name = self.get_resource_name()
        if ingress_name in active_resources_dict:
            active_resource = active_resources_dict[ingress_name]
            self.active_resource = active_resource
            logger.debug(f"Found active {ingress_name}")
        return active_resource

    def _update(self, k8s_client: K8sApiClient) -> bool:
        networking_v1_api: NetworkingV1Api = k8s_client.networking_v1_api
        ingress_name = self.get_resource_name()
        k8s_object: V1Ingress = self.get_k8s_object()
        namespace = self.get_namespace()

        logger.debug("Updating: {}".format(ingress_name))
        v1_ingress: V1Ingress = networking_v1_api.patch_namespaced_ingress(
            name=ingress_name,
            namespace=namespace,
            body=k8s_object,
            async_req=self.async_req,
            pretty=self.pretty,
        )
        # logger.debug("Updated:\n{}".format(pformat(v1_ingress.to_dict(), indent=2)))
        if v1_ingress.metadata.creation_timestamp is not None:
            logger.debug("Ingress Updated")
            self.active_resource = v1_ingress
            return True
        logger.error("Ingress could not be updated")
        return False

    def _delete(self, k8s_client: K8sApiClient) -> bool:
        networking_v1_api: NetworkingV1Api = k8s_client.networking_v1_api
        ingress_name = self.get_resource_name()
        namespace = self.get_namespace()

        logger.debug("Deleting: {}".format(ingress_name))
        self.active_resource = None
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_status.py
        delete_status: V1Status = networking_v1_api.delete_namespaced_ingress(
            name=ingress_name,
            namespace=namespace,
            async_req=self.async_req,
            pretty=self.pretty,
        )
        logger.debug("delete_status: {}".format(delete_status.status))
        if delete_status.status == "Success":
            logger.debug("Ingress Deleted")
            return True
        logger.error("Ingress could not be deleted")
        return False



from pathlib import Path
from typing import Optional, Any, Dict, List
from typing_extensions import Literal

from phi.aws.api_client import AwsApiClient
from phi.aws.resource.base import AwsResource
from phi.aws.resource.ec2.security_group import SecurityGroup
from phi.aws.resource.elasticache.subnet_group import CacheSubnetGroup
from phi.cli.console import print_info
from phi.utils.log import logger


class CacheCluster(AwsResource):
    """
    Reference:
    - https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/elasticache.html
    """

    resource_type: Optional[str] = "CacheCluster"
    service_name: str = "elasticache"

    # Name of the cluster.
    name: str
    # The node group (shard) identifier. This parameter is stored as a lowercase string.
    # If None, use the name as the cache_cluster_id
    # Constraints:
    #   A name must contain from 1 to 50 alphanumeric characters or hyphens.
    #   The first character must be a letter.
    #   A name cannot end with a hyphen or contain two consecutive hyphens.
    cache_cluster_id: Optional[str] = None
    # The name of the cache engine to be used for this cluster.
    engine: Literal["memcached", "redis"]

    # Compute and memory capacity of the nodes in the node group (shard).
    cache_node_type: str
    # The initial number of cache nodes that the cluster has.
    # For clusters running Redis, this value must be 1.
    # For clusters running Memcached, this value must be between 1 and 40.
    num_cache_nodes: int

    # The ID of the replication group to which this cluster should belong.
    # If this parameter is specified, the cluster is added to the specified replication group as a read replica;
    # otherwise, the cluster is a standalone primary that is not part of any replication group.
    replication_group_id: Optional[str] = None
    # Specifies whether the nodes in this Memcached cluster are created in a single Availability Zone or
    # created across multiple Availability Zones in the cluster's region.
    # This parameter is only supported for Memcached clusters.
    az_mode: Optional[Literal["single-az", "cross-az"]] = None
    # The EC2 Availability Zone in which the cluster is created.
    # All nodes belonging to this cluster are placed in the preferred Availability Zone. If you want to create your
    # nodes across multiple Availability Zones, use PreferredAvailabilityZones .
    # Default: System chosen Availability Zone.
    preferred_availability_zone: Optional[str] = None
    # A list of the Availability Zones in which cache nodes are created. The order of the zones is not important.
    # This option is only supported on Memcached.
    preferred_availability_zones: Optional[List[str]] = None
    # The version number of the cache engine to be used for this cluster.
    engine_version: Optional[str] = None
    cache_parameter_group_name: Optional[str] = None

    # The name of the subnet group to be used for the cluster.
    cache_subnet_group_name: Optional[str] = None
    # If cache_subnet_group_name is None,
    # Read the cache_subnet_group_name from cache_subnet_group
    cache_subnet_group: Optional[CacheSubnetGroup] = None

    # A list of security group names to associate with this cluster.
    # Use this parameter only when you are creating a cluster outside of an Amazon Virtual Private Cloud (Amazon VPC).
    cache_security_group_names: Optional[List[str]] = None
    # One or more VPC security groups associated with the cluster.
    # Use this parameter only when you are creating a cluster in an Amazon Virtual Private Cloud (Amazon VPC).
    cache_security_group_ids: Optional[List[str]] = None
    # If cache_security_group_ids is None
    # Read the security_group_id from cache_security_groups
    cache_security_groups: Optional[List[SecurityGroup]] = None

    tags: Optional[List[Dict[str, str]]] = None
    snapshot_arns: Optional[List[str]] = None
    snapshot_name: Optional[str] = None
    preferred_maintenance_window: Optional[str] = None
    # The version number of the cache engine to be used for this cluster.
    port: Optional[int] = None
    notification_topic_arn: Optional[str] = None
    auto_minor_version_upgrade: Optional[bool] = None
    snapshot_retention_limit: Optional[int] = None
    snapshot_window: Optional[str] = None
    # The password used to access a password protected server.
    # Password constraints:
    # - Must be only printable ASCII characters.
    # - Must be at least 16 characters and no more than 128 characters in length.
    # - The only permitted printable special characters are !, &, #, $, ^, <, >, and -.
    # Other printable special characters cannot be used in the AUTH token.
    # - For more information, see AUTH password at http://redis.io/commands/AUTH.
    # Provide AUTH_TOKEN here or as AUTH_TOKEN in secrets_file
    auth_token: Optional[str] = None
    outpost_mode: Optional[Literal["single-outpost", "cross-outpost"]] = None
    preferred_outpost_arn: Optional[str] = None
    preferred_outpost_arns: Optional[List[str]] = None
    log_delivery_configurations: Optional[List[Dict[str, Any]]] = None
    transit_encryption_enabled: Optional[bool] = None
    network_type: Optional[Literal["ipv4", "ipv6", "dual_stack"]] = None
    ip_discovery: Optional[Literal["ipv4", "ipv6"]] = None

    # The user-supplied name of a final cluster snapshot
    final_snapshot_identifier: Optional[str] = None

    # Read secrets from a file in yaml format
    secrets_file: Optional[Path] = None

    def get_cache_cluster_id(self):
        return self.cache_cluster_id or self.name

    def get_auth_token(self) -> Optional[str]:
        auth_token = self.auth_token
        if auth_token is None and self.secrets_file is not None:
            # read from secrets_file
            secret_data = self.get_secret_file_data()
            if secret_data is not None:
                auth_token = secret_data.get("AUTH_TOKEN", auth_token)
        return auth_token

    def _create(self, aws_client: AwsApiClient) -> bool:
        """Creates the CacheCluster

        Args:
            aws_client: The AwsApiClient for the current cluster
        """
        print_info(f"Creating {self.get_resource_type()}: {self.get_resource_name()}")

        # create a dict of args which are not null, otherwise aws type validation fails
        not_null_args: Dict[str, Any] = {}

        # Get the CacheSubnetGroupName
        cache_subnet_group_name = self.cache_subnet_group_name
        if cache_subnet_group_name is None and self.cache_subnet_group is not None:
            cache_subnet_group_name = self.cache_subnet_group.name
            logger.debug(f"Using CacheSubnetGroup: {cache_subnet_group_name}")
        if cache_subnet_group_name is not None:
            not_null_args["CacheSubnetGroupName"] = cache_subnet_group_name

        cache_security_group_ids = self.cache_security_group_ids
        if cache_security_group_ids is None and self.cache_security_groups is not None:
            sg_ids = []
            for sg in self.cache_security_groups:
                sg_id = sg.get_security_group_id(aws_client)
                if sg_id is not None:
                    sg_ids.append(sg_id)
            if len(sg_ids) > 0:
                cache_security_group_ids = sg_ids
                logger.debug(f"Using SecurityGroups: {cache_security_group_ids}")
        if cache_security_group_ids is not None:
            not_null_args["SecurityGroupIds"] = cache_security_group_ids

        if self.replication_group_id is not None:
            not_null_args["ReplicationGroupId"] = self.replication_group_id
        if self.az_mode is not None:
            not_null_args["AZMode"] = self.az_mode
        if self.preferred_availability_zone is not None:
            not_null_args["PreferredAvailabilityZone"] = self.preferred_availability_zone
        if self.preferred_availability_zones is not None:
            not_null_args["PreferredAvailabilityZones"] = self.preferred_availability_zones
        if self.num_cache_nodes is not None:
            not_null_args["NumCacheNodes"] = self.num_cache_nodes
        if self.cache_node_type is not None:
            not_null_args["CacheNodeType"] = self.cache_node_type
        if self.engine is not None:
            not_null_args["Engine"] = self.engine
        if self.engine_version is not None:
            not_null_args["EngineVersion"] = self.engine_version
        if self.cache_parameter_group_name is not None:
            not_null_args["CacheParameterGroupName"] = self.cache_parameter_group_name
        if self.cache_security_group_names is not None:
            not_null_args["CacheSecurityGroupNames"] = self.cache_security_group_names
        if self.tags is not None:
            not_null_args["Tags"] = self.tags
        if self.snapshot_arns is not None:
            not_null_args["SnapshotArns"] = self.snapshot_arns
        if self.snapshot_name is not None:
            not_null_args["SnapshotName"] = self.snapshot_name
        if self.preferred_maintenance_window is not None:
            not_null_args["PreferredMaintenanceWindow"] = self.preferred_maintenance_window
        if self.port is not None:
            not_null_args["Port"] = self.port
        if self.notification_topic_arn is not None:
            not_null_args["NotificationTopicArn"] = self.notification_topic_arn
        if self.auto_minor_version_upgrade is not None:
            not_null_args["AutoMinorVersionUpgrade"] = self.auto_minor_version_upgrade
        if self.snapshot_retention_limit is not None:
            not_null_args["SnapshotRetentionLimit"] = self.snapshot_retention_limit
        if self.snapshot_window is not None:
            not_null_args["SnapshotWindow"] = self.snapshot_window
        if self.auth_token is not None:
            not_null_args["AuthToken"] = self.get_auth_token()
        if self.outpost_mode is not None:
            not_null_args["OutpostMode"] = self.outpost_mode
        if self.preferred_outpost_arn is not None:
            not_null_args["PreferredOutpostArn"] = self.preferred_outpost_arn
        if self.preferred_outpost_arns is not None:
            not_null_args["PreferredOutpostArns"] = self.preferred_outpost_arns
        if self.log_delivery_configurations is not None:
            not_null_args["LogDeliveryConfigurations"] = self.log_delivery_configurations
        if self.transit_encryption_enabled is not None:
            not_null_args["TransitEncryptionEnabled"] = self.transit_encryption_enabled
        if self.network_type is not None:
            not_null_args["NetworkType"] = self.network_type
        if self.ip_discovery is not None:
            not_null_args["IpDiscovery"] = self.ip_discovery

        # Create CacheCluster
        service_client = self.get_service_client(aws_client)
        try:
            create_response = service_client.create_cache_cluster(
                CacheClusterId=self.get_cache_cluster_id(),
                **not_null_args,
            )
            logger.debug(f"CacheCluster: {create_response}")
            resource_dict = create_response.get("CacheCluster", {})

            # Validate resource creation
            if resource_dict is not None:
                print_info(f"CacheCluster created: {self.get_cache_cluster_id()}")
                self.active_resource = create_response
                return True
        except Exception as e:
            logger.error(f"{self.get_resource_type()} could not be created.")
            logger.error(e)
        return False

    def post_create(self, aws_client: AwsApiClient) -> bool:
        # Wait for CacheCluster to be created
        if self.wait_for_create:
            try:
                print_info(f"Waiting for {self.get_resource_type()} to be active.")
                waiter = self.get_service_client(aws_client).get_waiter("cache_cluster_available")
                waiter.wait(
                    CacheClusterId=self.get_cache_cluster_id(),
                    WaiterConfig={
                        "Delay": self.waiter_delay,
                        "MaxAttempts": self.waiter_max_attempts,
                    },
                )
            except Exception as e:
                logger.error("Waiter failed.")
                logger.error(e)
        return True

    def _read(self, aws_client: AwsApiClient) -> Optional[Any]:
        """Returns the CacheCluster

        Args:
            aws_client: The AwsApiClient for the current cluster
        """
        logger.debug(f"Reading {self.get_resource_type()}: {self.get_resource_name()}")

        from botocore.exceptions import ClientError

        service_client = self.get_service_client(aws_client)
        try:
            cache_cluster_id = self.get_cache_cluster_id()
            describe_response = service_client.describe_cache_clusters(CacheClusterId=cache_cluster_id)
            logger.debug(f"CacheCluster: {describe_response}")
            resource_list = describe_response.get("CacheClusters", None)

            if resource_list is not None and isinstance(resource_list, list):
                for resource in resource_list:
                    _cluster_identifier = resource.get("CacheClusterId", None)
                    if _cluster_identifier == cache_cluster_id:
                        self.active_resource = resource
                        break
        except ClientError as ce:
            logger.debug(f"ClientError: {ce}")
        except Exception as e:
            logger.error(f"Error reading {self.get_resource_type()}.")
            logger.error(e)
        return self.active_resource

    def _delete(self, aws_client: AwsApiClient) -> bool:
        """Deletes the CacheCluster

        Args:
            aws_client: The AwsApiClient for the current cluster
        """
        print_info(f"Deleting {self.get_resource_type()}: {self.get_resource_name()}")

        # create a dict of args which are not null, otherwise aws type validation fails
        not_null_args: Dict[str, Any] = {}
        if self.final_snapshot_identifier:
            not_null_args["FinalSnapshotIdentifier"] = self.final_snapshot_identifier

        service_client = self.get_service_client(aws_client)
        self.active_resource = None
        try:
            delete_response = service_client.delete_cache_cluster(
                CacheClusterId=self.get_cache_cluster_id(),
                **not_null_args,
            )
            logger.debug(f"CacheCluster: {delete_response}")
            return True
        except Exception as e:
            logger.error(f"{self.get_resource_type()} could not be deleted.")
            logger.error("Please try again or delete resources manually.")
            logger.error(e)
        return False

    def post_delete(self, aws_client: AwsApiClient) -> bool:
        # Wait for CacheCluster to be deleted
        if self.wait_for_delete:
            try:
                print_info(f"Waiting for {self.get_resource_type()} to be deleted.")
                waiter = self.get_service_client(aws_client).get_waiter("cache_cluster_deleted")
                waiter.wait(
                    CacheClusterId=self.get_cache_cluster_id(),
                    WaiterConfig={
                        "Delay": self.waiter_delay,
                        "MaxAttempts": self.waiter_max_attempts,
                    },
                )
            except Exception as e:
                logger.error("Waiter failed.")
                logger.error(e)
        return True

    def get_cache_endpoint(self, aws_client: Optional[AwsApiClient] = None) -> Optional[str]:
        """Returns the CacheCluster endpoint

        Args:
            aws_client: The AwsApiClient for the current cluster
        """
        cache_endpoint = None
        try:
            client: AwsApiClient = aws_client or self.get_aws_client()
            cache_cluster_id = self.get_cache_cluster_id()
            describe_response = self.get_service_client(client).describe_cache_clusters(CacheClusterId=cache_cluster_id)
            logger.debug(f"CacheCluster: {describe_response}")
            resource_list = describe_response.get("CacheClusters", None)

            if resource_list is not None and isinstance(resource_list, list):
                for resource in resource_list:
                    _cluster_identifier = resource.get("CacheClusterId", None)
                    if _cluster_identifier == cache_cluster_id:
                        cache_endpoint = resource.get("ConfigurationEndpoint", {}).get("Address", None)
                        break
        except Exception as e:
            logger.error(f"Error reading {self.get_resource_type()}.")
            logger.error(e)
        return cache_endpoint

    def get_cache_port(self, aws_client: Optional[AwsApiClient] = None) -> Optional[int]:
        """Returns the CacheCluster port

        Args:
            aws_client: The AwsApiClient for the current cluster
        """
        cache_port = None
        try:
            client: AwsApiClient = aws_client or self.get_aws_client()
            cache_cluster_id = self.get_cache_cluster_id()
            describe_response = self.get_service_client(client).describe_cache_clusters(CacheClusterId=cache_cluster_id)
            logger.debug(f"CacheCluster: {describe_response}")
            resource_list = describe_response.get("CacheClusters", None)

            if resource_list is not None and isinstance(resource_list, list):
                for resource in resource_list:
                    _cluster_identifier = resource.get("CacheClusterId", None)
                    if _cluster_identifier == cache_cluster_id:
                        cache_port = resource.get("ConfigurationEndpoint", {}).get("Port", None)
                        break
        except Exception as e:
            logger.error(f"Error reading {self.get_resource_type()}.")
            logger.error(e)
        return cache_port



from typing import Optional, Union, Dict, List, Iterator

from httpx import Response

from phi.api.api import api, invalid_response
from phi.api.routes import ApiRoutes
from phi.api.schemas.ai import (
    ConversationType,
    ConversationClient,
    ConversationCreateResponse,
)
from phi.api.schemas.user import UserSchema
from phi.llm.schemas import Function, Message
from phi.utils.log import logger


def conversation_create(
    user: UserSchema,
    conversation_type: ConversationType = ConversationType.RAG,
    functions: Optional[Dict[str, Function]] = None,
) -> Optional[ConversationCreateResponse]:
    logger.debug("--o-o-- Creating Conversation")
    with api.AuthenticatedClient() as api_client:
        try:
            r: Response = api_client.post(
                ApiRoutes.AI_CONVERSATION_CREATE,
                json={
                    "user": user.model_dump(include={"id_user", "email"}),
                    "conversation": {
                        "type": conversation_type,
                        "client": ConversationClient.CLI,
                        "functions": {
                            k: v.model_dump(include={"name", "description", "parameters"}) for k, v in functions.items()
                        }
                        if functions is not None
                        else None,
                    },
                },
            )
            if invalid_response(r):
                return None

            response_json: Union[Dict, List] = r.json()
            if response_json is None:
                return None

            new_conversation = ConversationCreateResponse.model_validate(response_json)
            if new_conversation is not None:
                return new_conversation
        except Exception as e:
            logger.debug(f"Could not create conversation: {e}")
    return None


def conversation_chat(
    user: UserSchema,
    conversation_id: int,
    message: Message,
    conversation_type: ConversationType = ConversationType.RAG,
    functions: Optional[Dict[str, Function]] = None,
    stream: bool = True,
) -> Optional[Iterator[str]]:
    with api.AuthenticatedClient() as api_client:
        if stream:
            logger.debug("--o-o-- Streaming Conversation Chat")
            try:
                with api_client.stream(
                    "POST",
                    ApiRoutes.AI_CONVERSATION_CHAT,
                    json={
                        "user": user.model_dump(include={"id_user", "email"}),
                        "conversation": {
                            "id": conversation_id,
                            "message": message.model_dump(exclude_none=True),
                            "type": conversation_type,
                            "client": ConversationClient.CLI,
                            "functions": {
                                k: v.model_dump(include={"name", "description", "parameters"})
                                for k, v in functions.items()
                            }
                            if functions is not None
                            else None,
                            "stream": stream,
                        },
                    },
                ) as streaming_resp:
                    for chunk in streaming_resp.iter_text():
                        yield chunk
            except Exception as e:
                logger.debug(f"Failed conversation chat: {e}")
        else:
            logger.debug("--o-o-- Conversation Chat")
            try:
                resp: Response = api_client.post(
                    ApiRoutes.AI_CONVERSATION_CHAT,
                    json={
                        "user": user.model_dump(include={"id_user", "email"}),
                        "conversation": {
                            "id": conversation_id,
                            "message": message.model_dump(exclude_none=True),
                            "type": conversation_type,
                            "client": ConversationClient.CLI,
                            "functions": {
                                k: v.model_dump(include={"name", "description", "parameters"})
                                for k, v in functions.items()
                            }
                            if functions is not None
                            else None,
                            "stream": stream,
                        },
                    },
                )
                if invalid_response(resp):
                    return None

                response_json: Optional[Dict] = resp.json()
                if response_json is None:
                    return None

                yield response_json.get("response")
            except Exception as e:
                logger.debug(f"Failed conversation chat: {e}")
    return None


# import os
# import base64
# import contextlib
# import wsproto
#
# class ConnectionClosed(Exception):
#     logger.debug("Connection closed")
#     pass
#
#
# class WebsocketConnection:
#     def __init__(self, network_steam):
#         self._ws_connection = wsproto.Connection(wsproto.ConnectionType.CLIENT)
#         self._network_stream = network_steam
#         self._events = []
#
#     async def send(self, text):
#         """
#         Send a text frame over the websocket connection.
#         """
#         try:
#             event = wsproto.events.TextMessage(text)
#             data = self._ws_connection.send(event)
#             await self._network_stream.write(data)
#         except Exception as e:
#             logger.debug(f"Failed to send: {e}")
#             logger.info("Connection closed, please start chat again.")
#             exit(1)
#
#     async def recv(self):
#         """
#         Receive the next text frame from the websocket connection.
#         """
#         try:
#             while not self._events:
#                 data = await self._network_stream.read(max_bytes=4096)
#                 self._ws_connection.receive_data(data)
#                 self._events = list(self._ws_connection.events())
#
#             event = self._events.pop(0)
#             if isinstance(event, wsproto.events.TextMessage):
#                 return event.data
#             elif isinstance(event, wsproto.events.CloseConnection):
#                 raise ConnectionClosed()
#         except Exception as e:
#             logger.debug(f"Failed to receive: {e}")
#             logger.info("Connection closed, please start chat again.")
#             exit(1)
#
#
# @contextlib.asynccontextmanager
# async def ai_ws_connect(
#     user: UserSchema,
#     conversation_id: int,
#     conversation_type: ConversationType = ConversationType.RAG,
#     stream: bool = True,
# ):
#     async with api.AuthenticatedAsyncClient() as api_client:
#         headers = {
#             "connection": "upgrade",
#             "upgrade": "websocket",
#             "sec-websocket-key": base64.b64encode(os.urandom(16)),
#             "sec-websocket-version": "13",
#             "X-PHIDATA-USER-ID": f"{user.id_user}",
#             "X-PHIDATA-CONVERSATION-ID": f"{conversation_id}",
#             "X-PHIDATA-CONVERSATION-TYPE": conversation_type.value,
#             "X-PHIDATA-CONVERSATION-STREAM": "true" if stream else "false",
#         }
#         headers.update(api.authenticated_headers)
#         logger.debug(f"Connecting to {ApiRoutes.AI_CONVERSATION_CHAT_WS}. Headers: {headers}")
#         async with api_client.stream(
#             "GET",
#             ApiRoutes.AI_CONVERSATION_CHAT_WS,
#             headers=headers,  # type: ignore
#         ) as response:
#             network_steam = response.extensions["network_stream"]
#             yield WebsocketConnection(network_steam)
#



from typing import List, Optional, Dict, Any, Union, Tuple

from pydantic import Field, field_validator, FieldValidationInfo

from phi.app.group import AppGroup
from phi.resource.group import ResourceGroup
from phi.k8s.app.base import K8sApp
from phi.k8s.app.context import K8sBuildContext
from phi.k8s.api_client import K8sApiClient
from phi.k8s.create.base import CreateK8sResource
from phi.k8s.resource.base import K8sResource
from phi.infra.resources import InfraResources
from phi.utils.log import logger


class K8sResources(InfraResources):
    apps: Optional[List[Union[K8sApp, AppGroup]]] = None
    resources: Optional[List[Union[K8sResource, CreateK8sResource, ResourceGroup]]] = None

    # K8s namespace to use
    namespace: str = "default"
    # K8s context to use
    context: Optional[str] = Field(None, validate_default=True)
    # Service account to use
    service_account_name: Optional[str] = None
    # Common labels to add to all resources
    common_labels: Optional[Dict[str, str]] = None
    # Path to kubeconfig file
    kubeconfig: Optional[str] = Field(None, validate_default=True)
    # Get context and kubeconfig from an EksCluster
    eks_cluster: Optional[Any] = None

    # -*- Cached Data
    _api_client: Optional[K8sApiClient] = None

    @property
    def k8s_client(self) -> K8sApiClient:
        if self._api_client is None:
            self._api_client = K8sApiClient(context=self.context, kubeconfig_path=self.kubeconfig)
        return self._api_client

    @field_validator("context", mode="before")
    def update_context(cls, context, info: FieldValidationInfo):
        if context is not None:
            return context

        # If context is not provided, then get it from eks_cluster
        eks_cluster = info.data.get("eks_cluster", None)
        if eks_cluster is not None:
            from phi.aws.resource.eks.cluster import EksCluster

            if not isinstance(eks_cluster, EksCluster):
                raise TypeError("eks_cluster not of type EksCluster")
            return eks_cluster.get_kubeconfig_context_name()
        return context

    @field_validator("kubeconfig", mode="before")
    def update_kubeconfig(cls, kubeconfig, info: FieldValidationInfo):
        if kubeconfig is not None:
            return kubeconfig

        # If kubeconfig is not provided, then get it from eks_cluster
        eks_cluster = info.data.get("eks_cluster", None)
        if eks_cluster is not None:
            from phi.aws.resource.eks.cluster import EksCluster

            if not isinstance(eks_cluster, EksCluster):
                raise TypeError("eks_cluster not of type EksCluster")
            return eks_cluster.kubeconfig_path
        return kubeconfig

    def create_resources(
        self,
        group_filter: Optional[str] = None,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        dry_run: Optional[bool] = False,
        auto_confirm: Optional[bool] = False,
        force: Optional[bool] = None,
    ) -> Tuple[int, int]:
        from phi.cli.console import print_info, print_heading, confirm_yes_no
        from phi.k8s.resource.types import K8sResourceInstallOrder

        logger.debug("-*- Creating K8sResources")
        # Build a list of K8sResources to create
        resources_to_create: List[K8sResource] = []
        if self.resources is not None:
            for r in self.resources:
                if isinstance(r, ResourceGroup):
                    resources_from_resource_group = r.get_resources()
                    if len(resources_from_resource_group) > 0:
                        for resource_from_resource_group in resources_from_resource_group:
                            if isinstance(resource_from_resource_group, K8sResource):
                                resource_from_resource_group.set_workspace_settings(
                                    workspace_settings=self.workspace_settings
                                )
                                if resource_from_resource_group.group is None and self.name is not None:
                                    resource_from_resource_group.group = self.name
                                if resource_from_resource_group.should_create(
                                    group_filter=group_filter,
                                    name_filter=name_filter,
                                    type_filter=type_filter,
                                ):
                                    resources_to_create.append(resource_from_resource_group)
                            elif isinstance(resource_from_resource_group, CreateK8sResource):
                                _k8s_resource = resource_from_resource_group.create()
                                if _k8s_resource is not None:
                                    _k8s_resource.set_workspace_settings(workspace_settings=self.workspace_settings)
                                    if _k8s_resource.group is None and self.name is not None:
                                        _k8s_resource.group = self.name
                                    if _k8s_resource.should_create(
                                        group_filter=group_filter,
                                        name_filter=name_filter,
                                        type_filter=type_filter,
                                    ):
                                        resources_to_create.append(_k8s_resource)
                elif isinstance(r, K8sResource):
                    r.set_workspace_settings(workspace_settings=self.workspace_settings)
                    if r.group is None and self.name is not None:
                        r.group = self.name
                    if r.should_create(
                        group_filter=group_filter,
                        name_filter=name_filter,
                        type_filter=type_filter,
                    ):
                        resources_to_create.append(r)
                elif isinstance(r, CreateK8sResource):
                    _k8s_resource = r.create()
                    if _k8s_resource is not None:
                        _k8s_resource.set_workspace_settings(workspace_settings=self.workspace_settings)
                        if _k8s_resource.group is None and self.name is not None:
                            _k8s_resource.group = self.name
                        if _k8s_resource.should_create(
                            group_filter=group_filter,
                            name_filter=name_filter,
                            type_filter=type_filter,
                        ):
                            resources_to_create.append(_k8s_resource)

        # Build a list of K8sApps to create
        apps_to_create: List[K8sApp] = []
        if self.apps is not None:
            for app in self.apps:
                if isinstance(app, AppGroup):
                    apps_from_app_group = app.get_apps()
                    if len(apps_from_app_group) > 0:
                        for app_from_app_group in apps_from_app_group:
                            if isinstance(app_from_app_group, K8sApp):
                                if app_from_app_group.group is None and self.name is not None:
                                    app_from_app_group.group = self.name
                                if app_from_app_group.should_create(group_filter=group_filter):
                                    apps_to_create.append(app_from_app_group)
                elif isinstance(app, K8sApp):
                    if app.group is None and self.name is not None:
                        app.group = self.name
                    if app.should_create(group_filter=group_filter):
                        apps_to_create.append(app)

        # Get the list of K8sResources from the K8sApps
        if len(apps_to_create) > 0:
            logger.debug(f"Found {len(apps_to_create)} apps to create")
            for app in apps_to_create:
                app.set_workspace_settings(workspace_settings=self.workspace_settings)
                app_resources = app.get_resources(
                    build_context=K8sBuildContext(
                        namespace=self.namespace,
                        context=self.context,
                        service_account_name=self.service_account_name,
                        labels=self.common_labels,
                    )
                )
                if len(app_resources) > 0:
                    for app_resource in app_resources:
                        if isinstance(app_resource, K8sResource) and app_resource.should_create(
                            group_filter=group_filter, name_filter=name_filter, type_filter=type_filter
                        ):
                            resources_to_create.append(app_resource)
                        elif isinstance(app_resource, CreateK8sResource):
                            _k8s_resource = app_resource.create()
                            if _k8s_resource is not None and _k8s_resource.should_create(
                                group_filter=group_filter, name_filter=name_filter, type_filter=type_filter
                            ):
                                resources_to_create.append(_k8s_resource)

        # Sort the K8sResources in install order
        resources_to_create.sort(key=lambda x: K8sResourceInstallOrder.get(x.__class__.__name__, 5000))

        # Deduplicate K8sResources
        deduped_resources_to_create: List[K8sResource] = []
        for r in resources_to_create:
            if r not in deduped_resources_to_create:
                deduped_resources_to_create.append(r)

        # Implement dependency sorting
        final_k8s_resources: List[K8sResource] = []
        logger.debug("-*- Building K8sResources dependency graph")
        for k8s_resource in deduped_resources_to_create:
            # Logic to follow if resource has dependencies
            if k8s_resource.depends_on is not None:
                # Add the dependencies before the resource itself
                for dep in k8s_resource.depends_on:
                    if isinstance(dep, K8sResource):
                        if dep not in final_k8s_resources:
                            logger.debug(f"-*- Adding {dep.name}, dependency of {k8s_resource.name}")
                            final_k8s_resources.append(dep)

                # Add the resource to be created after its dependencies
                if k8s_resource not in final_k8s_resources:
                    logger.debug(f"-*- Adding {k8s_resource.name}")
                    final_k8s_resources.append(k8s_resource)
            else:
                # Add the resource to be created if it has no dependencies
                if k8s_resource not in final_k8s_resources:
                    logger.debug(f"-*- Adding {k8s_resource.name}")
                    final_k8s_resources.append(k8s_resource)

        # Track the total number of K8sResources to create for validation
        num_resources_to_create: int = len(final_k8s_resources)
        num_resources_created: int = 0
        if num_resources_to_create == 0:
            return 0, 0

        if dry_run:
            print_heading("--**- K8s resources to create:")
            for resource in final_k8s_resources:
                print_info(f"  -+-> {resource.get_resource_type()}: {resource.get_resource_name()}")
            print_info("")
            print_info(f"Total {num_resources_to_create} resources")
            return 0, 0

        # Validate resources to be created
        if not auto_confirm:
            print_heading("\n--**-- Confirm resources to create:")
            for resource in final_k8s_resources:
                print_info(f"  -+-> {resource.get_resource_type()}: {resource.get_resource_name()}")
            print_info("")
            print_info(f"Total {num_resources_to_create} resources")
            confirm = confirm_yes_no("\nConfirm deploy")
            if not confirm:
                print_info("-*-")
                print_info("-*- Skipping deploy")
                print_info("-*-")
                return 0, 0

        for resource in final_k8s_resources:
            print_info(f"\n-==+==- {resource.get_resource_type()}: {resource.get_resource_name()}")
            if force is True:
                resource.force = True
            # logger.debug(resource)
            try:
                _resource_created = resource.create(k8s_client=self.k8s_client)
                if _resource_created:
                    num_resources_created += 1
                else:
                    if self.workspace_settings is not None and not self.workspace_settings.continue_on_create_failure:
                        return num_resources_created, num_resources_to_create
            except Exception as e:
                logger.error(f"Failed to create {resource.get_resource_type()}: {resource.get_resource_name()}")
                logger.error(e)
                logger.error("Please fix and try again...")

        print_heading(f"\n--**-- Resources created: {num_resources_created}/{num_resources_to_create}")
        if num_resources_to_create != num_resources_created:
            logger.error(
                f"Resources created: {num_resources_created} do not match resources required: {num_resources_to_create}"
            )  # noqa: E501
        return num_resources_created, num_resources_to_create

    def delete_resources(
        self,
        group_filter: Optional[str] = None,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        dry_run: Optional[bool] = False,
        auto_confirm: Optional[bool] = False,
        force: Optional[bool] = None,
    ) -> Tuple[int, int]:
        from phi.cli.console import print_info, print_heading, confirm_yes_no
        from phi.k8s.resource.types import K8sResourceInstallOrder

        logger.debug("-*- Deleting K8sResources")
        # Build a list of K8sResources to delete
        resources_to_delete: List[K8sResource] = []
        if self.resources is not None:
            for r in self.resources:
                if isinstance(r, ResourceGroup):
                    resources_from_resource_group = r.get_resources()
                    if len(resources_from_resource_group) > 0:
                        for resource_from_resource_group in resources_from_resource_group:
                            if isinstance(resource_from_resource_group, K8sResource):
                                resource_from_resource_group.set_workspace_settings(
                                    workspace_settings=self.workspace_settings
                                )
                                if resource_from_resource_group.group is None and self.name is not None:
                                    resource_from_resource_group.group = self.name
                                if resource_from_resource_group.should_delete(
                                    group_filter=group_filter,
                                    name_filter=name_filter,
                                    type_filter=type_filter,
                                ):
                                    resources_to_delete.append(resource_from_resource_group)
                            elif isinstance(resource_from_resource_group, CreateK8sResource):
                                _k8s_resource = resource_from_resource_group.create()
                                if _k8s_resource is not None:
                                    _k8s_resource.set_workspace_settings(workspace_settings=self.workspace_settings)
                                    if _k8s_resource.group is None and self.name is not None:
                                        _k8s_resource.group = self.name
                                    if _k8s_resource.should_delete(
                                        group_filter=group_filter,
                                        name_filter=name_filter,
                                        type_filter=type_filter,
                                    ):
                                        resources_to_delete.append(_k8s_resource)
                elif isinstance(r, K8sResource):
                    r.set_workspace_settings(workspace_settings=self.workspace_settings)
                    if r.group is None and self.name is not None:
                        r.group = self.name
                    if r.should_delete(
                        group_filter=group_filter,
                        name_filter=name_filter,
                        type_filter=type_filter,
                    ):
                        resources_to_delete.append(r)
                elif isinstance(r, CreateK8sResource):
                    _k8s_resource = r.create()
                    if _k8s_resource is not None:
                        _k8s_resource.set_workspace_settings(workspace_settings=self.workspace_settings)
                        if _k8s_resource.group is None and self.name is not None:
                            _k8s_resource.group = self.name
                        if _k8s_resource.should_delete(
                            group_filter=group_filter,
                            name_filter=name_filter,
                            type_filter=type_filter,
                        ):
                            resources_to_delete.append(_k8s_resource)

        # Build a list of K8sApps to delete
        apps_to_delete: List[K8sApp] = []
        if self.apps is not None:
            for app in self.apps:
                if isinstance(app, AppGroup):
                    apps_from_app_group = app.get_apps()
                    if len(apps_from_app_group) > 0:
                        for app_from_app_group in apps_from_app_group:
                            if isinstance(app_from_app_group, K8sApp):
                                if app_from_app_group.group is None and self.name is not None:
                                    app_from_app_group.group = self.name
                                if app_from_app_group.should_delete(group_filter=group_filter):
                                    apps_to_delete.append(app_from_app_group)
                elif isinstance(app, K8sApp):
                    if app.group is None and self.name is not None:
                        app.group = self.name
                    if app.should_delete(group_filter=group_filter):
                        apps_to_delete.append(app)

        # Get the list of K8sResources from the K8sApps
        if len(apps_to_delete) > 0:
            logger.debug(f"Found {len(apps_to_delete)} apps to delete")
            for app in apps_to_delete:
                app.set_workspace_settings(workspace_settings=self.workspace_settings)
                app_resources = app.get_resources(
                    build_context=K8sBuildContext(
                        namespace=self.namespace,
                        context=self.context,
                        service_account_name=self.service_account_name,
                        labels=self.common_labels,
                    )
                )
                if len(app_resources) > 0:
                    for app_resource in app_resources:
                        if isinstance(app_resource, K8sResource) and app_resource.should_delete(
                            group_filter=group_filter, name_filter=name_filter, type_filter=type_filter
                        ):
                            resources_to_delete.append(app_resource)
                        elif isinstance(app_resource, CreateK8sResource):
                            _k8s_resource = app_resource.create()
                            if _k8s_resource is not None and _k8s_resource.should_delete(
                                group_filter=group_filter, name_filter=name_filter, type_filter=type_filter
                            ):
                                resources_to_delete.append(_k8s_resource)

        # Sort the K8sResources in install order
        resources_to_delete.sort(key=lambda x: K8sResourceInstallOrder.get(x.__class__.__name__, 5000), reverse=True)

        # Deduplicate K8sResources
        deduped_resources_to_delete: List[K8sResource] = []
        for r in resources_to_delete:
            if r not in deduped_resources_to_delete:
                deduped_resources_to_delete.append(r)

        # Implement dependency sorting
        final_k8s_resources: List[K8sResource] = []
        logger.debug("-*- Building K8sResources dependency graph")
        for k8s_resource in deduped_resources_to_delete:
            # Logic to follow if resource has dependencies
            if k8s_resource.depends_on is not None:
                # 1. Reverse the order of dependencies
                k8s_resource.depends_on.reverse()

                # 2. Remove the dependencies if they are already added to the final_k8s_resources
                for dep in k8s_resource.depends_on:
                    if dep in final_k8s_resources:
                        logger.debug(f"-*- Removing {dep.name}, dependency of {k8s_resource.name}")
                        final_k8s_resources.remove(dep)

                # 3. Add the resource to be deleted before its dependencies
                if k8s_resource not in final_k8s_resources:
                    logger.debug(f"-*- Adding {k8s_resource.name}")
                    final_k8s_resources.append(k8s_resource)

                # 4. Add the dependencies back in reverse order
                for dep in k8s_resource.depends_on:
                    if isinstance(dep, K8sResource):
                        if dep not in final_k8s_resources:
                            logger.debug(f"-*- Adding {dep.name}, dependency of {k8s_resource.name}")
                            final_k8s_resources.append(dep)
            else:
                # Add the resource to be deleted if it has no dependencies
                if k8s_resource not in final_k8s_resources:
                    logger.debug(f"-*- Adding {k8s_resource.name}")
                    final_k8s_resources.append(k8s_resource)

        # Track the total number of K8sResources to delete for validation
        num_resources_to_delete: int = len(final_k8s_resources)
        num_resources_deleted: int = 0
        if num_resources_to_delete == 0:
            return 0, 0

        if dry_run:
            print_heading("--**- K8s resources to delete:")
            for resource in final_k8s_resources:
                print_info(f"  -+-> {resource.get_resource_type()}: {resource.get_resource_name()}")
            print_info("")
            print_info(f"Total {num_resources_to_delete} resources")
            return 0, 0

        # Validate resources to be deleted
        if not auto_confirm:
            print_heading("\n--**-- Confirm resources to delete:")
            for resource in final_k8s_resources:
                print_info(f"  -+-> {resource.get_resource_type()}: {resource.get_resource_name()}")
            print_info("")
            print_info(f"Total {num_resources_to_delete} resources")
            confirm = confirm_yes_no("\nConfirm delete")
            if not confirm:
                print_info("-*-")
                print_info("-*- Skipping delete")
                print_info("-*-")
                return 0, 0

        for resource in final_k8s_resources:
            print_info(f"\n-==+==- {resource.get_resource_type()}: {resource.get_resource_name()}")
            if force is True:
                resource.force = True
            # logger.debug(resource)
            try:
                _resource_deleted = resource.delete(k8s_client=self.k8s_client)
                if _resource_deleted:
                    num_resources_deleted += 1
                else:
                    if self.workspace_settings is not None and not self.workspace_settings.continue_on_delete_failure:
                        return num_resources_deleted, num_resources_to_delete
            except Exception as e:
                logger.error(f"Failed to delete {resource.get_resource_type()}: {resource.get_resource_name()}")
                logger.error(e)
                logger.error("Please fix and try again...")

        print_heading(f"\n--**-- Resources deleted: {num_resources_deleted}/{num_resources_to_delete}")
        if num_resources_to_delete != num_resources_deleted:
            logger.error(
                f"Resources deleted: {num_resources_deleted} do not match resources required: {num_resources_to_delete}"
            )  # noqa: E501
        return num_resources_deleted, num_resources_to_delete

    def update_resources(
        self,
        group_filter: Optional[str] = None,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        dry_run: Optional[bool] = False,
        auto_confirm: Optional[bool] = False,
        force: Optional[bool] = None,
    ) -> Tuple[int, int]:
        from phi.cli.console import print_info, print_heading, confirm_yes_no
        from phi.k8s.resource.types import K8sResourceInstallOrder

        logger.debug("-*- Updating K8sResources")

        # Build a list of K8sResources to update
        resources_to_update: List[K8sResource] = []
        if self.resources is not None:
            for r in self.resources:
                if isinstance(r, ResourceGroup):
                    resources_from_resource_group = r.get_resources()
                    if len(resources_from_resource_group) > 0:
                        for resource_from_resource_group in resources_from_resource_group:
                            if isinstance(resource_from_resource_group, K8sResource):
                                resource_from_resource_group.set_workspace_settings(
                                    workspace_settings=self.workspace_settings
                                )
                                if resource_from_resource_group.group is None and self.name is not None:
                                    resource_from_resource_group.group = self.name
                                if resource_from_resource_group.should_update(
                                    group_filter=group_filter,
                                    name_filter=name_filter,
                                    type_filter=type_filter,
                                ):
                                    resources_to_update.append(resource_from_resource_group)
                            elif isinstance(resource_from_resource_group, CreateK8sResource):
                                _k8s_resource = resource_from_resource_group.create()
                                if _k8s_resource is not None:
                                    _k8s_resource.set_workspace_settings(workspace_settings=self.workspace_settings)
                                    if _k8s_resource.group is None and self.name is not None:
                                        _k8s_resource.group = self.name
                                    if _k8s_resource.should_update(
                                        group_filter=group_filter,
                                        name_filter=name_filter,
                                        type_filter=type_filter,
                                    ):
                                        resources_to_update.append(_k8s_resource)
                elif isinstance(r, K8sResource):
                    r.set_workspace_settings(workspace_settings=self.workspace_settings)
                    if r.group is None and self.name is not None:
                        r.group = self.name
                    if r.should_update(
                        group_filter=group_filter,
                        name_filter=name_filter,
                        type_filter=type_filter,
                    ):
                        resources_to_update.append(r)
                elif isinstance(r, CreateK8sResource):
                    _k8s_resource = r.create()
                    if _k8s_resource is not None:
                        _k8s_resource.set_workspace_settings(workspace_settings=self.workspace_settings)
                        if _k8s_resource.group is None and self.name is not None:
                            _k8s_resource.group = self.name
                        if _k8s_resource.should_update(
                            group_filter=group_filter,
                            name_filter=name_filter,
                            type_filter=type_filter,
                        ):
                            resources_to_update.append(_k8s_resource)

        # Build a list of K8sApps to update
        apps_to_update: List[K8sApp] = []
        if self.apps is not None:
            for app in self.apps:
                if isinstance(app, AppGroup):
                    apps_from_app_group = app.get_apps()
                    if len(apps_from_app_group) > 0:
                        for app_from_app_group in apps_from_app_group:
                            if isinstance(app_from_app_group, K8sApp):
                                if app_from_app_group.group is None and self.name is not None:
                                    app_from_app_group.group = self.name
                                if app_from_app_group.should_update(group_filter=group_filter):
                                    apps_to_update.append(app_from_app_group)
                elif isinstance(app, K8sApp):
                    if app.group is None and self.name is not None:
                        app.group = self.name
                    if app.should_update(group_filter=group_filter):
                        apps_to_update.append(app)

        # Get the list of K8sResources from the K8sApps
        if len(apps_to_update) > 0:
            logger.debug(f"Found {len(apps_to_update)} apps to update")
            for app in apps_to_update:
                app.set_workspace_settings(workspace_settings=self.workspace_settings)
                app_resources = app.get_resources(
                    build_context=K8sBuildContext(
                        namespace=self.namespace,
                        context=self.context,
                        service_account_name=self.service_account_name,
                        labels=self.common_labels,
                    )
                )
                if len(app_resources) > 0:
                    for app_resource in app_resources:
                        if isinstance(app_resource, K8sResource) and app_resource.should_update(
                            group_filter=group_filter, name_filter=name_filter, type_filter=type_filter
                        ):
                            resources_to_update.append(app_resource)
                        elif isinstance(app_resource, CreateK8sResource):
                            _k8s_resource = app_resource.create()
                            if _k8s_resource is not None and _k8s_resource.should_update(
                                group_filter=group_filter, name_filter=name_filter, type_filter=type_filter
                            ):
                                resources_to_update.append(_k8s_resource)

        # Sort the K8sResources in install order
        resources_to_update.sort(key=lambda x: K8sResourceInstallOrder.get(x.__class__.__name__, 5000), reverse=True)

        # Deduplicate K8sResources
        deduped_resources_to_update: List[K8sResource] = []
        for r in resources_to_update:
            if r not in deduped_resources_to_update:
                deduped_resources_to_update.append(r)

        # Implement dependency sorting
        final_k8s_resources: List[K8sResource] = []
        logger.debug("-*- Building K8sResources dependency graph")
        for k8s_resource in deduped_resources_to_update:
            # Logic to follow if resource has dependencies
            if k8s_resource.depends_on is not None:
                # Add the dependencies before the resource itself
                for dep in k8s_resource.depends_on:
                    if isinstance(dep, K8sResource):
                        if dep not in final_k8s_resources:
                            logger.debug(f"-*- Adding {dep.name}, dependency of {k8s_resource.name}")
                            final_k8s_resources.append(dep)

                # Add the resource to be created after its dependencies
                if k8s_resource not in final_k8s_resources:
                    logger.debug(f"-*- Adding {k8s_resource.name}")
                    final_k8s_resources.append(k8s_resource)
            else:
                # Add the resource to be created if it has no dependencies
                if k8s_resource not in final_k8s_resources:
                    logger.debug(f"-*- Adding {k8s_resource.name}")
                    final_k8s_resources.append(k8s_resource)

        # Track the total number of K8sResources to update for validation
        num_resources_to_update: int = len(final_k8s_resources)
        num_resources_updated: int = 0
        if num_resources_to_update == 0:
            return 0, 0

        if dry_run:
            print_heading("--**- K8s resources to update:")
            for resource in final_k8s_resources:
                print_info(f"  -+-> {resource.get_resource_type()}: {resource.get_resource_name()}")
            print_info("")
            print_info(f"Total {num_resources_to_update} resources")
            return 0, 0

        # Validate resources to be updated
        if not auto_confirm:
            print_heading("\n--**-- Confirm resources to update:")
            for resource in final_k8s_resources:
                print_info(f"  -+-> {resource.get_resource_type()}: {resource.get_resource_name()}")
            print_info("")
            print_info(f"Total {num_resources_to_update} resources")
            confirm = confirm_yes_no("\nConfirm patch")
            if not confirm:
                print_info("-*-")
                print_info("-*- Skipping patch")
                print_info("-*-")
                return 0, 0

        for resource in final_k8s_resources:
            print_info(f"\n-==+==- {resource.get_resource_type()}: {resource.get_resource_name()}")
            if force is True:
                resource.force = True
            # logger.debug(resource)
            try:
                _resource_updated = resource.update(k8s_client=self.k8s_client)
                if _resource_updated:
                    num_resources_updated += 1
                else:
                    if self.workspace_settings is not None and not self.workspace_settings.continue_on_patch_failure:
                        return num_resources_updated, num_resources_to_update
            except Exception as e:
                logger.error(f"Failed to update {resource.get_resource_type()}: {resource.get_resource_name()}")
                logger.error(e)
                logger.error("Please fix and try again...")

        print_heading(f"\n--**-- Resources updated: {num_resources_updated}/{num_resources_to_update}")
        if num_resources_to_update != num_resources_updated:
            logger.error(
                f"Resources updated: {num_resources_updated} do not match resources required: {num_resources_to_update}"
            )  # noqa: E501
        return num_resources_updated, num_resources_to_update

    def save_resources(
        self,
        group_filter: Optional[str] = None,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
    ) -> Tuple[int, int]:
        from phi.cli.console import print_info, print_heading
        from phi.k8s.resource.types import K8sResourceInstallOrder

        logger.debug("-*- Saving K8sResources")

        # Build a list of K8sResources to save
        resources_to_save: List[K8sResource] = []
        if self.resources is not None:
            for r in self.resources:
                if isinstance(r, ResourceGroup):
                    resources_from_resource_group = r.get_resources()
                    if len(resources_from_resource_group) > 0:
                        for resource_from_resource_group in resources_from_resource_group:
                            if isinstance(resource_from_resource_group, K8sResource):
                                resource_from_resource_group.env = self.env
                                resource_from_resource_group.set_workspace_settings(
                                    workspace_settings=self.workspace_settings
                                )
                                if resource_from_resource_group.group is None and self.name is not None:
                                    resource_from_resource_group.group = self.name
                                if resource_from_resource_group.should_create(
                                    group_filter=group_filter,
                                    name_filter=name_filter,
                                    type_filter=type_filter,
                                ):
                                    resources_to_save.append(resource_from_resource_group)
                            elif isinstance(resource_from_resource_group, CreateK8sResource):
                                _k8s_resource = resource_from_resource_group.save()
                                if _k8s_resource is not None:
                                    _k8s_resource.env = self.env
                                    _k8s_resource.set_workspace_settings(workspace_settings=self.workspace_settings)
                                    if _k8s_resource.group is None and self.name is not None:
                                        _k8s_resource.group = self.name
                                    if _k8s_resource.should_create(
                                        group_filter=group_filter,
                                        name_filter=name_filter,
                                        type_filter=type_filter,
                                    ):
                                        resources_to_save.append(_k8s_resource)
                elif isinstance(r, K8sResource):
                    r.env = self.env
                    r.set_workspace_settings(workspace_settings=self.workspace_settings)
                    if r.group is None and self.name is not None:
                        r.group = self.name
                    if r.should_create(
                        group_filter=group_filter,
                        name_filter=name_filter,
                        type_filter=type_filter,
                    ):
                        resources_to_save.append(r)
                elif isinstance(r, CreateK8sResource):
                    _k8s_resource = r.create()
                    if _k8s_resource is not None:
                        _k8s_resource.env = self.env
                        _k8s_resource.set_workspace_settings(workspace_settings=self.workspace_settings)
                        if _k8s_resource.group is None and self.name is not None:
                            _k8s_resource.group = self.name
                        if _k8s_resource.should_create(
                            group_filter=group_filter,
                            name_filter=name_filter,
                            type_filter=type_filter,
                        ):
                            resources_to_save.append(_k8s_resource)

        # Build a list of K8sApps to save
        apps_to_save: List[K8sApp] = []
        if self.apps is not None:
            for app in self.apps:
                if isinstance(app, AppGroup):
                    apps_from_app_group = app.get_apps()
                    if len(apps_from_app_group) > 0:
                        for app_from_app_group in apps_from_app_group:
                            if isinstance(app_from_app_group, K8sApp):
                                if app_from_app_group.group is None and self.name is not None:
                                    app_from_app_group.group = self.name
                                if app_from_app_group.should_create(group_filter=group_filter):
                                    apps_to_save.append(app_from_app_group)
                elif isinstance(app, K8sApp):
                    if app.group is None and self.name is not None:
                        app.group = self.name
                    if app.should_create(group_filter=group_filter):
                        apps_to_save.append(app)

        # Get the list of K8sResources from the K8sApps
        if len(apps_to_save) > 0:
            logger.debug(f"Found {len(apps_to_save)} apps to save")
            for app in apps_to_save:
                app.set_workspace_settings(workspace_settings=self.workspace_settings)
                app_resources = app.get_resources(
                    build_context=K8sBuildContext(
                        namespace=self.namespace,
                        context=self.context,
                        service_account_name=self.service_account_name,
                        labels=self.common_labels,
                    )
                )
                if len(app_resources) > 0:
                    for app_resource in app_resources:
                        if isinstance(app_resource, K8sResource) and app_resource.should_create(
                            group_filter=group_filter, name_filter=name_filter, type_filter=type_filter
                        ):
                            resources_to_save.append(app_resource)
                        elif isinstance(app_resource, CreateK8sResource):
                            _k8s_resource = app_resource.save()
                            if _k8s_resource is not None and _k8s_resource.should_create(
                                group_filter=group_filter, name_filter=name_filter, type_filter=type_filter
                            ):
                                resources_to_save.append(_k8s_resource)

        # Sort the K8sResources in install order
        resources_to_save.sort(key=lambda x: K8sResourceInstallOrder.get(x.__class__.__name__, 5000))

        # Deduplicate K8sResources
        deduped_resources_to_save: List[K8sResource] = []
        for r in resources_to_save:
            if r not in deduped_resources_to_save:
                deduped_resources_to_save.append(r)

        # Implement dependency sorting
        final_k8s_resources: List[K8sResource] = []
        logger.debug("-*- Building K8sResources dependency graph")
        for k8s_resource in deduped_resources_to_save:
            # Logic to follow if resource has dependencies
            if k8s_resource.depends_on is not None:
                # Add the dependencies before the resource itself
                for dep in k8s_resource.depends_on:
                    if isinstance(dep, K8sResource):
                        if dep not in final_k8s_resources:
                            logger.debug(f"-*- Adding {dep.name}, dependency of {k8s_resource.name}")
                            final_k8s_resources.append(dep)

                # Add the resource to be saved after its dependencies
                if k8s_resource not in final_k8s_resources:
                    logger.debug(f"-*- Adding {k8s_resource.name}")
                    final_k8s_resources.append(k8s_resource)
            else:
                # Add the resource to be saved if it has no dependencies
                if k8s_resource not in final_k8s_resources:
                    logger.debug(f"-*- Adding {k8s_resource.name}")
                    final_k8s_resources.append(k8s_resource)

        # Track the total number of K8sResources to save for validation
        num_resources_to_save: int = len(final_k8s_resources)
        num_resources_saved: int = 0
        if num_resources_to_save == 0:
            return 0, 0

        for resource in final_k8s_resources:
            print_info(f"\n-==+==- {resource.get_resource_type()}: {resource.get_resource_name()}")
            try:
                _resource_path = resource.save_manifests(default_flow_style=False)
                if _resource_path is not None:
                    print_info(f"Saved to: {_resource_path}")
                    num_resources_saved += 1
            except Exception as e:
                logger.error(f"Failed to save {resource.get_resource_type()}: {resource.get_resource_name()}")
                logger.error(e)
                logger.error("Please fix and try again...")

        print_heading(f"\n--**-- Resources saved: {num_resources_saved}/{num_resources_to_save}")
        if num_resources_to_save != num_resources_saved:
            logger.error(
                f"Resources saved: {num_resources_saved} do not match resources required: {num_resources_to_save}"
            )  # noqa: E501
        return num_resources_saved, num_resources_to_save



from typing import Dict, List, Optional, Union
from typing_extensions import Literal

from phi.k8s.create.core.v1.container import CreateContainer
from phi.k8s.create.core.v1.volume import CreateVolume
from phi.k8s.create.base import CreateK8sResource
from phi.k8s.enums.api_version import ApiVersion
from phi.k8s.enums.kind import Kind
from phi.k8s.enums.restart_policy import RestartPolicy
from phi.k8s.resource.apps.v1.deployment import (
    Deployment,
    DeploymentSpec,
    LabelSelector,
    PodTemplateSpec,
)
from phi.k8s.resource.core.v1.container import Container
from phi.k8s.resource.core.v1.pod_spec import PodSpec
from phi.k8s.resource.core.v1.volume import Volume
from phi.k8s.resource.core.v1.topology_spread_constraints import (
    TopologySpreadConstraint,
)
from phi.k8s.create.common.labels import create_component_labels_dict
from phi.k8s.resource.meta.v1.object_meta import ObjectMeta


class CreateDeployment(CreateK8sResource):
    deploy_name: str
    pod_name: str
    app_name: str
    namespace: Optional[str] = None
    service_account_name: Optional[str] = None
    replicas: Optional[int] = 1
    containers: List[CreateContainer]
    init_containers: Optional[List[CreateContainer]] = None
    pod_node_selector: Optional[Dict[str, str]] = None
    restart_policy: RestartPolicy = RestartPolicy.ALWAYS
    termination_grace_period_seconds: Optional[int] = None
    volumes: Optional[List[CreateVolume]] = None
    labels: Optional[Dict[str, str]] = None
    pod_annotations: Optional[Dict[str, str]] = None
    topology_spread_key: Optional[str] = None
    topology_spread_max_skew: Optional[int] = None
    topology_spread_when_unsatisfiable: Optional[Union[str, Literal["DoNotSchedule", "ScheduleAnyway"]]] = None
    # If True, recreate the resource on update
    # Used for deployments with EBS volumes
    recreate_on_update: bool = False

    def _create(self) -> Deployment:
        """Creates the Deployment resource"""

        deploy_name = self.deploy_name
        # logger.debug(f"Init Deployment resource: {deploy_name}")

        deploy_labels = create_component_labels_dict(
            component_name=deploy_name,
            app_name=self.app_name,
            labels=self.labels,
        )

        pod_name = self.pod_name
        pod_labels = create_component_labels_dict(
            component_name=pod_name,
            app_name=self.app_name,
            labels=self.labels,
        )

        containers: List[Container] = []
        for cc in self.containers:
            container = cc.create()
            if container is not None and isinstance(container, Container):
                containers.append(container)

        init_containers: Optional[List[Container]] = None
        if self.init_containers is not None:
            init_containers = []
            for ic in self.init_containers:
                _init_container = ic.create()
                if _init_container is not None and isinstance(_init_container, Container):
                    init_containers.append(_init_container)

        topology_spread_constraints: Optional[List[TopologySpreadConstraint]] = None
        if self.topology_spread_key is not None:
            topology_spread_constraints = [
                TopologySpreadConstraint(
                    topology_key=self.topology_spread_key,
                    max_skew=self.topology_spread_max_skew,
                    when_unsatisfiable=self.topology_spread_when_unsatisfiable,
                    label_selector=LabelSelector(match_labels=pod_labels),
                )
            ]

        volumes: Optional[List[Volume]] = None
        if self.volumes:
            volumes = []
            for cv in self.volumes:
                volume = cv.create()
                if volume and isinstance(volume, Volume):
                    volumes.append(volume)

        deployment = Deployment(
            name=deploy_name,
            api_version=ApiVersion.APPS_V1,
            kind=Kind.DEPLOYMENT,
            metadata=ObjectMeta(
                name=deploy_name,
                namespace=self.namespace,
                labels=deploy_labels,
            ),
            spec=DeploymentSpec(
                replicas=self.replicas,
                selector=LabelSelector(match_labels=pod_labels),
                template=PodTemplateSpec(
                    # TODO: fix this
                    metadata=ObjectMeta(
                        name=pod_name,
                        namespace=self.namespace,
                        labels=pod_labels,
                        annotations=self.pod_annotations,
                    ),
                    spec=PodSpec(
                        init_containers=init_containers,
                        node_selector=self.pod_node_selector,
                        service_account_name=self.service_account_name,
                        restart_policy=self.restart_policy,
                        containers=containers,
                        termination_grace_period_seconds=self.termination_grace_period_seconds,
                        topology_spread_constraints=topology_spread_constraints,
                        volumes=volumes,
                    ),
                ),
            ),
            recreate_on_update=self.recreate_on_update,
        )

        # logger.debug(
        #     f"Deployment {deploy_name}:\n{deployment.json(exclude_defaults=True, indent=2)}"
        # )
        return deployment



from typing import Optional, Union, List

from phi.k8s.app.superset.base import SupersetBase


class SupersetWorker(SupersetBase):
    # -*- App Name
    name: str = "superset-worker"

    # Command for the container
    command: Optional[Union[str, List[str]]] = "worker"



from time import sleep
from typing import Optional, Any, Dict, Union, List

from phi.docker.api_client import DockerApiClient
from phi.docker.resource.base import DockerResource
from phi.cli.console import print_info
from phi.utils.log import logger


class DockerContainerMount(DockerResource):
    resource_type: str = "ContainerMount"

    target: str
    source: str
    type: str = "volume"
    read_only: bool = False
    labels: Optional[Dict[str, Any]] = None


class DockerContainer(DockerResource):
    resource_type: str = "Container"

    # image (str) – The image to run.
    image: Optional[str] = None
    # command (str or list) – The command to run in the container.
    command: Optional[Union[str, List]] = None
    # auto_remove (bool) – enable auto-removal of the container when the container’s process exits.
    auto_remove: bool = True
    # detach (bool) – Run container in the background and return a Container object.
    detach: bool = True
    # entrypoint (str or list) – The entrypoint for the container.
    entrypoint: Optional[Union[str, List]] = None
    # environment (dict or list) – Environment variables to set inside the container
    environment: Optional[Union[Dict[str, Any], List]] = None
    # group_add (list) – List of additional group names and/or IDs that the container process will run as.
    group_add: Optional[List[Any]] = None
    # healthcheck (dict) – Specify a test to perform to check that the container is healthy.
    healthcheck: Optional[Dict[str, Any]] = None
    # hostname (str) – Optional hostname for the container.
    hostname: Optional[str] = None
    # labels (dict or list) – A dictionary of name-value labels
    # e.g. {"label1": "value1", "label2": "value2"})
    # or a list of names of labels to set with empty values (e.g. ["label1", "label2"])
    labels: Optional[Dict[str, Any]] = None
    # mounts (list) – Specification for mounts to be added to the container.
    # More powerful alternative to volumes.
    # Each item in the list is a DockerContainerMount object which is
    # then converted to a docker.types.Mount object.
    mounts: Optional[List[DockerContainerMount]] = None
    # network (str) – Name of the network this container will be connected to at creation time
    network: Optional[str] = None
    # network_disabled (bool) – Disable networking.
    network_disabled: Optional[str] = None
    # network_mode (str) One of:
    # bridge - Create a new network stack for the container on on the bridge network.
    # none - No networking for this container.
    # container:<name|id> - Reuse another container’s network stack.
    # host - Use the host network stack. This mode is incompatible with ports.
    # network_mode is incompatible with network.
    network_mode: Optional[str] = None
    # Platform in the format os[/arch[/variant]].
    platform: Optional[str] = None
    # ports (dict) – Ports to bind inside the container.
    # The keys of the dictionary are the ports to bind inside the container,
    # either as an integer or a string in the form port/protocol, where the protocol is either tcp, udp.
    #
    # The values of the dictionary are the corresponding ports to open on the host, which can be either:
    #   - The port number, as an integer.
    #       For example, {'2222/tcp': 3333} will expose port 2222 inside the container
    #       as port 3333 on the host.
    #   - None, to assign a random host port. For example, {'2222/tcp': None}.
    #   - A tuple of (address, port) if you want to specify the host interface.
    #       For example, {'1111/tcp': ('127.0.0.1', 1111)}.
    #   - A list of integers, if you want to bind multiple host ports to a single container port.
    #       For example, {'1111/tcp': [1234, 4567]}.
    ports: Optional[Dict[str, Any]] = None
    # remove (bool) – Remove the container when it has finished running. Default: False.
    remove: Optional[bool] = None
    # Restart the container when it exits. Configured as a dictionary with keys:
    # Name: One of on-failure, or always.
    # MaximumRetryCount: Number of times to restart the container on failure.
    # For example: {"Name": "on-failure", "MaximumRetryCount": 5}
    restart_policy: Optional[Dict[str, Any]] = None
    # stdin_open (bool) – Keep STDIN open even if not attached.
    stdin_open: Optional[bool] = None
    # stdout (bool) – Return logs from STDOUT when detach=False. Default: True.
    stdout: Optional[bool] = None
    # stderr (bool) – Return logs from STDERR when detach=False. Default: False.
    stderr: Optional[bool] = None
    # tty (bool) – Allocate a pseudo-TTY.
    tty: Optional[bool] = None
    # user (str or int) – Username or UID to run commands as inside the container.
    user: Optional[Union[str, int]] = None
    # volumes (dict or list) –
    # A dictionary to configure volumes mounted inside the container.
    # The key is either the host path or a volume name, and the value is a dictionary with the keys:
    # bind - The path to mount the volume inside the container
    # mode - Either rw to mount the volume read/write, or ro to mount it read-only.
    # For example:
    # {
    #   '/home/user1/': {'bind': '/mnt/vol2', 'mode': 'rw'},
    #   '/var/www': {'bind': '/mnt/vol1', 'mode': 'ro'}
    # }
    volumes: Optional[Union[Dict[str, Any], List]] = None
    # working_dir (str) – Path to the working directory.
    working_dir: Optional[str] = None
    devices: Optional[list] = None

    # Data provided by the resource running on the docker client
    container_status: Optional[str] = None

    def run_container(self, docker_client: DockerApiClient) -> Optional[Any]:
        from docker import DockerClient
        from docker.errors import ImageNotFound, APIError
        from rich.progress import Progress

        print_info("Starting container: {}".format(self.name))
        # logger.debug()(
        #     "Args: {}".format(
        #         self.json(indent=2, exclude_unset=True, exclude_none=True)
        #     )
        # )
        try:
            _api_client: DockerClient = docker_client.api_client
            with Progress(transient=True) as progress:
                task = progress.add_task("Downloading Image", total=None)  # noqa: F841
                container_object = _api_client.containers.run(
                    name=self.name,
                    image=self.image,
                    command=self.command,
                    auto_remove=self.auto_remove,
                    detach=self.detach,
                    entrypoint=self.entrypoint,
                    environment=self.environment,
                    group_add=self.group_add,
                    healthcheck=self.healthcheck,
                    hostname=self.hostname,
                    labels=self.labels,
                    mounts=self.mounts,
                    network=self.network,
                    network_disabled=self.network_disabled,
                    network_mode=self.network_mode,
                    platform=self.platform,
                    ports=self.ports,
                    remove=self.remove,
                    restart_policy=self.restart_policy,
                    stdin_open=self.stdin_open,
                    stdout=self.stdout,
                    stderr=self.stderr,
                    tty=self.tty,
                    user=self.user,
                    volumes=self.volumes,
                    working_dir=self.working_dir,
                    devices=self.devices,
                )
                return container_object
        except ImageNotFound as img_error:
            logger.error(f"Image {self.image} not found. Explanation: {img_error.explanation}")
            raise
        except APIError as api_err:
            logger.error(f"APIError: {api_err.explanation}")
            raise
        except Exception:
            raise

    def _create(self, docker_client: DockerApiClient) -> bool:
        """Creates the Container

        Args:
            docker_client: The DockerApiClient for the current cluster
        """
        from docker.models.containers import Container

        logger.debug("Creating: {}".format(self.get_resource_name()))
        container_object: Optional[Container] = self._read(docker_client)

        # Delete the container if it exists
        if container_object is not None:
            print_info(f"Deleting container {container_object.name}")
            self._delete(docker_client)

        try:
            container_object = self.run_container(docker_client)
            if container_object is not None:
                logger.debug("Container Created: {}".format(container_object.name))
            else:
                logger.debug("Container could not be created")
        except Exception:
            raise

        # By this step the container should be created
        # Validate that the container is running
        logger.debug("Validating container is created...")
        if container_object is not None:
            container_object.reload()
            self.container_status: str = container_object.status
            print_info("Container Status: {}".format(self.container_status))

            if self.container_status == "running":
                logger.debug("Container is running")
                return True
            elif self.container_status == "created":
                from rich.progress import Progress

                with Progress(transient=True) as progress:
                    task = progress.add_task("Waiting for container to start", total=None)  # noqa: F841
                    while self.container_status != "created":
                        logger.debug(f"Container Status: {self.container_status}, trying again in 1 seconds")
                        sleep(1)
                        container_object.reload()
                        self.container_status = container_object.status
                    logger.debug(f"Container Status: {self.container_status}")

            if self.container_status in ("running", "created"):
                logger.debug("Container Created")
                self.active_resource = container_object
                return True

        logger.debug("Container not found")
        return False

    def _read(self, docker_client: DockerApiClient) -> Optional[Any]:
        """Returns a Container object if the container is active

        Args:
            docker_client: The DockerApiClient for the current cluster
        """
        from docker import DockerClient
        from docker.models.containers import Container

        logger.debug("Reading: {}".format(self.get_resource_name()))
        container_name: Optional[str] = self.name
        try:
            _api_client: DockerClient = docker_client.api_client
            container_list: Optional[List[Container]] = _api_client.containers.list(
                all=True, filters={"name": container_name}
            )
            if container_list is not None:
                for container in container_list:
                    if container.name == container_name:
                        logger.debug(f"Container {container_name} exists")
                        self.active_resource = container
                        return container
        except Exception:
            logger.debug(f"Container {container_name} not found")
        return None

    def _update(self, docker_client: DockerApiClient) -> bool:
        """Updates the Container

        Args:
            docker_client: The DockerApiClient for the current cluster
        """
        logger.debug("Updating: {}".format(self.get_resource_name()))
        return self._create(docker_client=docker_client)

    def _delete(self, docker_client: DockerApiClient) -> bool:
        """Deletes the Container

        Args:
            docker_client: The DockerApiClient for the current cluster
        """
        from docker.models.containers import Container
        from docker.errors import NotFound

        logger.debug("Deleting: {}".format(self.get_resource_name()))
        container_name: Optional[str] = self.name
        container_object: Optional[Container] = self._read(docker_client)
        # Return True if there is no Container to delete
        if container_object is None:
            return True

        # Delete Container
        try:
            self.active_resource = None
            self.container_status = container_object.status
            logger.debug("Container Status: {}".format(self.container_status))
            logger.debug("Stopping Container: {}".format(container_name))
            container_object.stop()
            # If self.remove is set, then the container would be auto removed after being stopped
            # If self.remove is not set, we need to manually remove the container
            if not self.remove:
                logger.debug("Removing Container: {}".format(container_name))
                try:
                    container_object.remove()
                except Exception as remove_exc:
                    logger.debug(f"Could not remove container: {remove_exc}")
        except Exception as e:
            logger.exception("Error while deleting container: {}".format(e))

        # Validate that the Container is deleted
        logger.debug("Validating Container is deleted")
        try:
            logger.debug("Reloading container_object: {}".format(container_object))
            for i in range(10):
                container_object.reload()
                logger.debug("Waiting for NotFound Exception...")
                sleep(1)
        except NotFound:
            logger.debug("Got NotFound Exception, container is deleted")

        return True

    def is_active(self, docker_client: DockerApiClient) -> bool:
        """Returns True if the container is running on the docker cluster"""
        from docker.models.containers import Container

        container_object: Optional[Container] = self.read(docker_client=docker_client)
        if container_object is not None:
            # Check if container is stopped/paused
            status: str = container_object.status
            if status in ["exited", "paused"]:
                logger.debug(f"Container status: {status}")
                return False
            return True
        return False

    def create(self, docker_client: DockerApiClient) -> bool:
        # If self.force then always create container
        if not self.force:
            # If use_cache is True and container is active then return True
            if self.use_cache and self.is_active(docker_client=docker_client):
                print_info(f"{self.get_resource_type()}: {self.get_resource_name()} already exists")
                return True

        resource_created = self._create(docker_client=docker_client)
        if resource_created:
            print_info(f"{self.get_resource_type()}: {self.get_resource_name()} created")
            return True
        logger.error(f"Failed to create {self.get_resource_type()}: {self.get_resource_name()}")
        return False



"""Phidata Kubectl Cli

This is the entrypoint for the `phi k` commands.
"""
from pathlib import Path
from typing import Optional

import typer

from phi.cli.console import (
    print_info,
    log_config_not_available_msg,
    log_active_workspace_not_available,
    print_available_workspaces,
)
from phi.utils.log import logger, set_log_level_to_debug

k_cli = typer.Typer(
    name="k",
    short_help="Manage kubernetes resources",
    help="""\b
Use `phi k [COMMAND]` to save, get, update kubernetes resources.
Run `phi k [COMMAND] --help` for more info.
""",
    no_args_is_help=True,
    add_completion=False,
    invoke_without_command=True,
    options_metavar="",
    subcommand_metavar="[COMMAND] [OPTIONS]",
)


@k_cli.command(short_help="Save your K8s Resources")
def save(
    resource_filter: Optional[str] = typer.Argument(
        None,
        help="Resource filter. Format - ENV:GROUP:NAME:TYPE",
    ),
    env_filter: Optional[str] = typer.Option(None, "-e", "--env", metavar="", help="Filter the environment to deploy."),
    group_filter: Optional[str] = typer.Option(
        None, "-g", "--group", metavar="", help="Filter resources using group name."
    ),
    name_filter: Optional[str] = typer.Option(None, "-n", "--name", metavar="", help="Filter resource using name."),
    type_filter: Optional[str] = typer.Option(
        None,
        "-t",
        "--type",
        metavar="",
        help="Filter resource using type",
    ),
    print_debug_log: bool = typer.Option(
        False,
        "-d",
        "--debug",
        help="Print debug logs.",
    ),
):
    """
    Saves your k8s resources. Used to validate what is being deployed

    \b
    Examples:
    > `phi k save`      -> Save resources for the active workspace
    """
    if print_debug_log:
        set_log_level_to_debug()

    from phi.cli.config import PhiCliConfig
    from phi.workspace.config import WorkspaceConfig
    from phi.k8s.operator import save_resources
    from phi.utils.resource_filter import parse_k8s_resource_filter

    phi_config: Optional[PhiCliConfig] = PhiCliConfig.from_saved_config()
    if not phi_config:
        log_config_not_available_msg()
        return

    active_ws_config: Optional[WorkspaceConfig] = phi_config.get_active_ws_config()
    if active_ws_config is None:
        log_active_workspace_not_available()
        avl_ws = phi_config.available_ws
        if avl_ws:
            print_available_workspaces(avl_ws)
        return

    current_path: Path = Path(".").resolve()
    if active_ws_config.ws_root_path != current_path:
        ws_at_current_path = phi_config.get_ws_config_by_path(current_path)
        if ws_at_current_path is not None:
            print_info(
                f"Workspace at the current directory ({ws_at_current_path.ws_dir_name}) "
                + f"is not the Active Workspace ({active_ws_config.ws_dir_name})"
            )
            update_active_workspace = typer.confirm(
                f"Update active workspace to {ws_at_current_path.ws_dir_name}", default=True
            )
            if update_active_workspace:
                phi_config.active_ws_dir = ws_at_current_path.ws_dir_name
                active_ws_config = ws_at_current_path

    target_env: Optional[str] = None
    target_group: Optional[str] = None
    target_name: Optional[str] = None
    target_type: Optional[str] = None

    # derive env:infra:name:type:group from ws_filter
    if resource_filter is not None:
        if not isinstance(resource_filter, str):
            raise TypeError(f"Invalid resource_filter. Expected: str, Received: {type(resource_filter)}")
        (
            target_env,
            target_group,
            target_name,
            target_type,
        ) = parse_k8s_resource_filter(resource_filter)

    # derive env:infra:name:type:group from command options
    if target_env is None and env_filter is not None and isinstance(env_filter, str):
        target_env = env_filter
    if target_group is None and group_filter is not None and isinstance(group_filter, str):
        target_group = group_filter
    if target_name is None and name_filter is not None and isinstance(name_filter, str):
        target_name = name_filter
    if target_type is None and type_filter is not None and isinstance(type_filter, str):
        target_type = type_filter

    logger.debug("Processing workspace")
    logger.debug(f"\ttarget_env   : {target_env}")
    logger.debug(f"\ttarget_group : {target_group}")
    logger.debug(f"\ttarget_name  : {target_name}")
    logger.debug(f"\ttarget_type  : {target_type}")
    save_resources(
        phi_config=phi_config,
        ws_config=active_ws_config,
        target_env=target_env,
        target_group=target_group,
        target_name=target_name,
        target_type=target_type,
    )


# @app.command(short_help="Print your K8s Resources")
# def print(
#     refresh: bool = typer.Option(
#         False,
#         "-r",
#         "--refresh",
#         help="Refresh the workspace config, use this if you've just changed your phi-config.yaml",
#         show_default=True,
#     ),
#     type_filters: List[str] = typer.Option(
#         None, "-k", "--kind", help="Filter the K8s resources by kind"
#     ),
#     name_filters: List[str] = typer.Option(
#         None, "-n", "--name", help="Filter the K8s resources by name"
#     ),
# ):
#     """
#     Print your k8s resources so you know exactly what is being deploying
#
#     \b
#     Examples:
#     * `phi k print`      -> Print resources for the primary workspace
#     * `phi k print data` -> Print resources for the workspace named data
#     """
#
#     from phi import schemas
#     from phiterm.k8s import k8s_operator
#     from phiterm.conf.phi_conf import PhiConf
#
#     config: Optional[PhiConf] = PhiConf.get_saved_conf()
#     if not config:
#         conf_not_available_msg()
#         raise typer.Exit(1)
#
#     primary_ws: Optional[schemas.WorkspaceSchema] = config.primary_ws
#     if primary_ws is None:
#         primary_ws_not_available_msg()
#         raise typer.Exit(1)
#
#     k8s_operator.print_k8s_resources_as_yaml(
#         primary_ws, config, refresh, type_filters, name_filters
#     )
#
#
# @app.command(short_help="Apply your K8s Resources")
# def apply(
#     refresh: bool = typer.Option(
#         False,
#         "-r",
#         "--refresh",
#         help="Refresh the workspace config, use this if you've just changed your phi-config.yaml",
#         show_default=True,
#     ),
#     service_filters: List[str] = typer.Option(
#         None, "-s", "--svc", help="Filter the Services"
#     ),
#     type_filters: List[str] = typer.Option(
#         None, "-k", "--kind", help="Filter the K8s resources by kind"
#     ),
#     name_filters: List[str] = typer.Option(
#         None, "-n", "--name", help="Filter the K8s resources by name"
#     ),
# ):
#     """
#     Apply your k8s resources. You can filter the resources by services, kind or name
#
#     \b
#     Examples:
#     * `phi k apply`      -> Apply resources for the primary workspace
#     """
#
#     from phi import schemas
#     from phiterm.k8s import k8s_operator
#     from phiterm.conf.phi_conf import PhiConf
#
#     config: Optional[PhiConf] = PhiConf.get_saved_conf()
#     if not config:
#         conf_not_available_msg()
#         raise typer.Exit(1)
#
#     primary_ws: Optional[schemas.WorkspaceSchema] = config.primary_ws
#     if primary_ws is None:
#         primary_ws_not_available_msg()
#         raise typer.Exit(1)
#
#     k8s_operator.apply_k8s_resources(
#         primary_ws, config, refresh, service_filters, type_filters, name_filters
#     )
#
#
# @app.command(short_help="Get active K8s Objects")
# def get(
#     service_filters: List[str] = typer.Option(
#         None, "-s", "--svc", help="Filter the Services"
#     ),
#     type_filters: List[str] = typer.Option(
#         None, "-k", "--kind", help="Filter the K8s resources by kind"
#     ),
#     name_filters: List[str] = typer.Option(
#         None, "-n", "--name", help="Filter the K8s resources by name"
#     ),
# ):
#     """
#     Get active k8s resources.
#
#     \b
#     Examples:
#     * `phi k apply`      -> Get active resources for the primary workspace
#     """
#
#     from phi import schemas
#     from phiterm.k8s import k8s_operator
#     from phiterm.conf.phi_conf import PhiConf
#
#     config: Optional[PhiConf] = PhiConf.get_saved_conf()
#     if not config:
#         conf_not_available_msg()
#         raise typer.Exit(1)
#
#     primary_ws: Optional[schemas.WorkspaceSchema] = config.primary_ws
#     if primary_ws is None:
#         primary_ws_not_available_msg()
#         raise typer.Exit(1)
#
#     k8s_operator.print_active_k8s_resources(
#         primary_ws, config, service_filters, type_filters, name_filters
#     )
#
#
# if __name__ == "__main__":
#     app()



from pathlib import Path
from typing import Union, List, Iterator

from phi.document import Document
from phi.document.reader.text import TextReader
from phi.knowledge.base import KnowledgeBase


class TextKnowledgeBase(KnowledgeBase):
    path: Union[str, Path]
    formats: List[str] = [".doc", ".docx"]
    reader: TextReader = TextReader()

    @property
    def document_lists(self) -> Iterator[List[Document]]:
        """Iterate over text files and yield lists of documents.
        Each object yielded by the iterator is a list of documents.

        Returns:
            Iterator[List[Document]]: Iterator yielding list of documents
        """

        _file_path: Path = Path(self.path) if isinstance(self.path, str) else self.path

        if _file_path.exists() and _file_path.is_dir():
            for _file in _file_path.glob("**/*"):
                if _file.suffix in self.formats:
                    yield self.reader.read(path=_file)
        elif _file_path.exists() and _file_path.is_file() and _file_path.suffix == ".pdf":
            yield self.reader.read(path=_file_path)



from datetime import datetime
from typing import Optional, Dict, Any

from pydantic import BaseModel


class ConversationWorkspace(BaseModel):
    id_workspace: int
    ws_hash: Optional[str] = None


class ConversationEventCreate(BaseModel):
    """Data sent to API to create a new conversation event"""

    id_user: Optional[int] = None
    conversation_key: str
    conversation_data: Optional[Dict[str, Any]] = None
    event_type: str
    event_data: Optional[Dict[str, Any]] = None


class ConversationEventCreateResopnse(BaseModel):
    id_event: Optional[int] = None
    id_conversation: Optional[int] = None


class ConversationUpdate(BaseModel):
    """Data sent to API to update a conversation"""

    conversation_key: str
    conversation_data: Optional[Dict[str, Any]] = None


class ConversationSchema(BaseModel):
    """Schema for a conversation returned by API"""

    id_conversation: Optional[int] = None
    id_workspace: Optional[int] = None
    conversation_key: Optional[str] = None
    conversation_data: Optional[Dict[str, Any]] = None
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None



from typing import Dict, Optional

from kubernetes.client.models.v1_resource_requirements import V1ResourceRequirements

from phi.k8s.resource.base import K8sObject


class ResourceRequirements(K8sObject):
    """
    ResourceRequirements describes the compute resource requirements.
    - https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#resourcerequirements-v1-core
    """

    resource_type: str = "ResourceRequirements"

    # Limits describes the maximum amount of compute resources allowed
    limits: Optional[Dict[str, str]] = None
    # Requests describes the minimum amount of compute resources required.
    # If Requests is omitted for a container, it defaults to Limits if that is explicitly specified,
    # otherwise to an implementation-defined value.
    requests: Optional[Dict[str, str]] = None

    def get_k8s_object(self) -> V1ResourceRequirements:
        # Return a V1ResourceRequirements object
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_resource_requirements.py
        _v1_resource_requirements = V1ResourceRequirements(
            limits=self.limits,
            requests=self.requests,
        )
        return _v1_resource_requirements
